"""Contains Optislang class, which provides the Python API for the optiSLang app."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Mapping, Optional, Sequence, Tuple, Union

from importlib_metadata import version

from ansys.optislang.core import LOG
from ansys.optislang.core.project import Project
from ansys.optislang.core.tcp_osl_server import TcpOslServer

if TYPE_CHECKING:
    from ansys.optislang.core.logging import OslCustomAdapter
    from ansys.optislang.core.osl_server import OslServer


class Optislang:
    """Connects to the optiSLang app and provides an API to control it.

    For remote connection, it is assumed that optiSLang is already running
    on a remote (or local) host as a server. In this case, the host and port
    must be specified. Parameters related to the execution of the new optiSLang
    server are ignored.

    To run optiSLang locally, both the ``host`` and ``port`` parameters must
    be ``None``, which are their defaults. Other parameters can be optionally
    specified.

    Parameters
    ----------
    host : str, optional
        IPv4/v6 address or domain name on which optiSLang is running as a
        server. The default is ``None``.
    port : int, optional
        Port on which optiSLang is running as a server. The default is ``None``.
    executable : Union[str, pathlib.Path], optional
        Path to the optiSLang executable file to execute on a the local host.
        The default is ``None``. This parameter is ignored when ``host``
        and ``port`` parameters are specified.
    project_path : Union[str, pathlib.Path], optional
        Path to the optiSLang project file that a new local optiSLang server
        is to use. The default is ``None``. This parameter is ignored
        when ``host`` and ``port`` parameters are specified. Here is how
        this parameter is used:

        - If the project file exists, it is opened.
        - If the project file does not exist, a project is created in the specified path.
        - If the path is ``None``, a new project is created in the temporary directory.

    batch : bool, optional
        Determines whether to start optiSLang server in batch mode. Defaults to ``True``.
    port_range : Tuple[int, int], optional
        Defines the port range for optiSLang server. Defaults to ``None``.
    no_run : bool, optional
        Determines whether not to run the specified project when started in batch mode.
        Defaults to ``None``.

        .. note:: Only supported in batch mode.

    no_save : bool, optional
        Whether to save the specified project after all other actions are completed.
        The default is ``False``. This parameter is ignored when ``host`` and
        ``port`` parameters are specified.

        .. note:: Only supported in batch mode.

    force : bool, optional
        Determines whether to force opening/processing specified project when started in batch mode
        even if issues occur.
        Defaults to ``True``.

        .. note:: Only supported in batch mode.

    reset : bool, optional
        Determines whether to reset specified project after load.
        Defaults to ``False``.

        .. note:: Only supported in batch mode.

    auto_relocate : bool, optional
        Determines whether to automatically relocate missing file paths.
        Defaults to ``False``.

        .. note:: Only supported in batch mode.

    listener_id : str, optional
        Specific unique ID for the TCP listener. Defaults to ``None``.
    multi_listener : Iterable[Tuple[str, int, Optional[str]]], optional
        Multiple remote listeners (plain TCP/IP based) to be registered at optiSLang server.
        Each listener is a combination of host, port and (optionally) listener ID.
        Defaults to ``None``.
    ini_timeout : float, optional
        Time in seconds to connect to the optiSLang server. The default is ``20``.
    name : str, optional
        ID of the optiSLang instance.
    password : str, optional
        Server password. The default is ``None``. This parameter is used when
        communication with the server requires that the request contain a password.
    loglevel : str, optional
        Logging level. The options are:

        - ``CRITICAL``: Log errors that are fatal for the app.
        - ``ERROR``: Log errors that are fatal for some operation, but not for the app.
        - ``WARNING``: Log some oddities or potential problems.
        - ``INFO``: Log some useful information that the program works as expected.
        - ``DEBUG``: Log all information for use in debugging.

    shutdown_on_finished: bool, optional
        Whether to shut down when execution is finished and no listeners are registered.
        The default is ``True``. This parameter is ignored when ``host`` and
        ``port`` parameters are specified.

        .. note:: Only supported in batch mode.

    env_vars : Mapping[str, str], optional
        Additional environmental variables (key and value) for the optiSLang server process.
        Defaults to ``None``.
    import_project_properties_file : Union[str, pathlib.Path], optional
        Optional path to a project properties file to import. Defaults to ``None``.
    export_project_properties_file : Union[str, pathlib.Path], optional
        Optional path to a project properties file to export. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    import_placeholders_file : Union[str, pathlib.Path], optional
        Optional path to a placeholders file to import. Defaults to ``None``.
    export_placeholders_file : Union[str, pathlib.Path], optional
        Optional path to a placeholders file to export. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    output_file : Union[str, pathlib.Path], optional
        Optional path to an output file for writing project run results to. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    dump_project_state : Union[str, pathlib.Path], optional
        Optional path to a project state dump file to export. If a relative path is provided,
        it is considered to be relative to the project working directory. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    opx_project_definition_file : Union[str, pathlib.Path], optional
        Optional path to an OPX project definition file. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    additional_args : Iterable[str], optional
        Additional command line arguments used for execution of the optiSLang server process.
        Defaults to ``None``.

    Raises
    ------
    RuntimeError
        Raised when the connection to the optiSLang server cannot be established
        before the specified timeout.

    Examples
    --------
    Start and connect to the local optiSLang server, get the version of optiSLang
    in use, print the version, and shut down the server.

    >>> from ansys.optislang.core import Optislang
    >>> osl = Optislang()
    >>> osl_version = osl.get_osl_version_string()
    >>> print(osl_version)
    >>> osl.dispose()
    """

    def __init__(
        self,
        host: str = None,
        port: int = None,
        executable: Union[str, Path] = None,
        project_path: Union[str, Path] = None,
        batch: bool = True,
        port_range: Tuple[int, int] = None,
        no_run: bool = None,
        no_save: bool = False,
        force: bool = True,
        reset: bool = False,
        auto_relocate: bool = False,
        listener_id: str = None,
        multi_listener: Iterable[Tuple[str, int, Optional[str]]] = None,
        ini_timeout: Union[int, float] = 20,
        name: str = None,
        password: str = None,
        loglevel: str = None,
        shutdown_on_finished: bool = True,
        env_vars: Mapping[str, str] = None,
        import_project_properties_file: Union[str, Path] = None,
        export_project_properties_file: Union[str, Path] = None,
        import_placeholders_file: Union[str, Path] = None,
        export_placeholders_file: Union[str, Path] = None,
        output_file: Union[str, Path] = None,
        dump_project_state: Union[str, Path] = None,
        opx_project_definition_file: Union[str, Path] = None,
        additional_args: Iterable[str] = None,
    ) -> None:
        """Initialize a new instance of the ``Optislang`` class."""
        self.__host = host
        self.__port = port
        self.__executable = Path(executable) if executable is not None else None
        self.__project_path = Path(project_path) if project_path is not None else None
        self.__batch = batch
        self.__port_range = port_range
        self.__no_run = no_run
        self.__no_save = no_save
        self.__force = force
        self.__reset = reset
        self.__auto_relocate = auto_relocate
        self.__ini_timeout = ini_timeout
        self.__name = name
        self.__password = password
        self.__shutdown_on_finished = shutdown_on_finished
        self.__env_vars = env_vars
        self.__listener_id = listener_id
        self.__multi_listener = multi_listener
        self.__import_project_properties_file = import_project_properties_file
        self.__export_project_properties_file = export_project_properties_file
        self.__import_placeholders_file = import_placeholders_file
        self.__export_placeholders_file = export_placeholders_file
        self.__output_file = output_file
        self.__dump_project_state = dump_project_state
        self.__opx_project_definition_file = opx_project_definition_file
        self.__additional_args = additional_args
        self.__logger = LOG.add_instance_logger(self.name, self, loglevel)
        self.__osl_server: OslServer = self.__init_osl_server("tcp")
        project_uid = self.__osl_server.get_project_uid()
        self.__project = (
            Project(osl_server=self.__osl_server, uid=project_uid) if project_uid else None
        )

    def __init_osl_server(self, server_type: str) -> OslServer:
        """Initialize optiSLang server.

        Parameters
        ----------
        server_type : str, optional
            Type of the optiSLang server. The default is ``tcp``, in which case
            the plain TCP/IP communication protocol is used.

        Returns
        -------
        OslServer
            OptiSlang server object.

        Raises
        ------
        NotImplementedError
            Raised when the specified server type is not supported.
        """
        if server_type.lower() == "tcp":
            return TcpOslServer(
                host=self.__host,
                port=self.__port,
                executable=self.__executable,
                project_path=self.__project_path,
                no_save=self.__no_save,
                ini_timeout=self.__ini_timeout,
                password=self.__password,
                logger=self.log,
                shutdown_on_finished=self.__shutdown_on_finished,
                batch=self.__batch,
                port_range=self.__port_range,
                no_run=self.__no_run,
                force=self.__force,
                reset=self.__reset,
                auto_relocate=self.__auto_relocate,
                env_vars=self.__env_vars,
                import_project_properties_file=self.__import_project_properties_file,
                export_project_properties_file=self.__export_project_properties_file,
                import_placeholders_file=self.__import_placeholders_file,
                export_placeholders_file=self.__export_placeholders_file,
                output_file=self.__output_file,
                dump_project_state=self.__dump_project_state,
                opx_project_definition_file=self.__opx_project_definition_file,
                listener_id=self.__listener_id,
                multi_listener=self.__multi_listener,
                additional_args=self.__additional_args,
            )
        else:
            raise NotImplementedError(f'OptiSLang server of type "{server_type}" is not supported.')

    def __str__(self):
        """Return product name, version of optiSLang, and version of PyOptiSLang."""
        return (
            f"Product name: optiSLang\n"
            f"Version: {self.get_osl_version_string()}\n"
            f"PyOptiSLang: {version('ansys.optislang.core')}"
        )

    def __enter__(self):
        """Enter the context."""
        self.log.debug("Enter the context.")
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Exit the context.

        Disposes the Optislang instance.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        self.log.debug("Exit the context.")
        self.dispose()

    @property
    def name(self) -> str:
        """Unique ID of the optiSLang instance."""
        if not self.__name:
            if self.__host or self.__port:
                self.__name = f"optiSLang_{self.__host}:{self.__port}"
            else:
                self.__name = f"optiSLang_{id(self)}"
        return self.__name

    @property
    def log(self) -> OslCustomAdapter:
        """Instance logger."""
        return self.__logger

    @property
    def has_active_project(self) -> bool:
        """
        Whether a project is loaded.

        Returns
        -------
        bool
            ``True`` if a project is loaded, ``False`` otherwise.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.get_project_uid() is not None

    @property
    def project(self) -> Project:
        """Instance of the ``Project`` class.

        Returns
        -------
        Project
            Loaded project. If no project is loaded, ``None`` is returned.
        """
        return self.__project

    # close method doesn't work properly in optiSLang 2023 R1, Thus, it was commented out
    # def close(self) -> None:
    #     """Close the current project.

    #     Raises
    #     ------
    #     OslCommunicationError
    #         Raised when an error occurs while communicating with the server.
    #     OslCommandError
    #         Raised when a command or query fails.
    #     TimeoutError
    #         Raised when the timeout float value expires.
    #     """
    #     self.__osl_server.close()
    #     self.__project = None

    def dispose(self) -> None:
        """Close all local threads and unregister listeners.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        self.__osl_server.dispose()

    def get_osl_version_string(self) -> str:
        """Get the optiSLang version in use as a string.

        Returns
        -------
        str
            optiSLang version.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.get_osl_version_string()

    def get_osl_version(self) -> Tuple[Union[int, None], ...]:
        """Get the optiSLang version in use as a tuple.

        Returns
        -------
        tuple
            optiSLang version as tuple contains the
            major version, minor version, maintenance version, and revision.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.get_osl_version()

    def get_timeout(self) -> Union[float, None]:
        """Get the timeout value for executing commands.

        Returns
        -------
        timeout: Union[float, None]
            Timeout in seconds to perform commands. This value must be greater
            than zero or ``None``. The default is ``None``. Another function
            raises a timeout exception if the timeout value has elapsed before
            an operation has completed. If the timeout is ``None``, functions
            wait until they're finished, and no timeout exception is raised.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.get_timeout()

    def get_working_dir(self) -> Path:
        """Get the path to the optiSLang project's working directory.

        Returns
        -------
        pathlib.Path
            Path to the optiSLang project's working directory. If no project is loaded
            in optiSLang, ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.get_working_dir()

    def new(self) -> None:
        """Create a project.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        self.__osl_server.new()
        self.__project = Project(
            osl_server=self.__osl_server, uid=self.__osl_server.get_project_uid()
        )

    def open(
        self,
        file_path: Union[str, Path],
        force: bool = True,
        restore: bool = False,
        reset: bool = False,
    ) -> None:
        """Open a project.

        Parameters
        ----------
        file_path : Union[str, pathlib.Path]
            Path to the optiSLang project file to open.
        force : bool, optional
            Whether to force opening of the project even if non-critical errors occur.
            Non-critical errors include:

            - Timestamp of the (auto) save point is newer than the project timestamp.
            - Project (file) is incomplete.

        restore : bool, optional
            Whether to restore the project from the last (auto) save point (if present).
        reset : bool, optional
            Whether to reset the project after loading it.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        self.__osl_server.open(file_path=file_path, force=force, restore=restore, reset=reset)
        self.__project = Project(
            osl_server=self.__osl_server, uid=self.__osl_server.get_project_uid()
        )

    def reset(self) -> None:
        """Reset the project.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        self.__osl_server.reset()

    def run_python_script(
        self,
        script: str,
        args: Union[Sequence[object], None] = None,
    ) -> Tuple[str, str]:
        """Load a Python script in a project context and run it.

        Parameters
        ----------
        script : str
            Python commands to execute on the server.
        args : Sequence[object], None, optional
            Sequence of arguments used in the Python script. The default
            is ``None``.

        Returns
        -------
        Tuple[str, str]
            STDOUT and STDERR from the executed Python script.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.run_python_script(script, args)

    def run_python_file(
        self,
        file_path: Union[str, Path],
        args: Union[Sequence[object], None] = None,
    ) -> Tuple[str, str]:
        """Read a Python script from a file, load it in a project context, and run it.

        Parameters
        ----------
        file_path : Union[str, pathlib.Path]
            Path to the Python script file with the content to execute on the server.
        args : Sequence[object], None, optional
            Sequence of arguments to use in the Python script. The default is ``None``.

        Returns
        -------
        Tuple[str, str]
            STDOUT and STDERR from the executed Python script.

        Raises
        ------
        FileNotFoundError
            Raised when the specified Python script file does not exist.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.run_python_file(file_path, args)

    def save(self) -> None:
        """Save changes to the project data and settings.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.save()

    def save_as(
        self,
        file_path: Union[str, Path],
        force: bool = True,
        restore: bool = False,
        reset: bool = False,
    ) -> None:
        """Save and open the project at a new location.

        Parameters
        ----------
        file_path : Union[str, pathlib.Path]
            Path for saving the new project file to.
        force : bool, optional
            Whether to force opening of the project even if non-critical errors occur.
            Non-critical errors include:

            - Timestamp of the (auto) save point is newer than the project timestamp.
            - Project (file) is incomplete.

        restore : bool, optional
            Whether to restore the project from the last (auto) save point (if present).
        reset : bool, optional
            Whether to reset the project after loading it.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.save_as(
            file_path=file_path, force=force, restore=restore, reset=reset
        )

    def save_copy(self, file_path: Union[str, Path]) -> None:
        """Save a copy of the project to a specified location.

        Parameters
        ----------
        file_path : Union[str, pathlib.Path]
            Path for saving the copy of the project file to.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        self.__osl_server.save_copy(file_path)

    def set_timeout(self, timeout: Union[float, None] = None) -> None:
        """Set the timeout value for the executing commands.

        Parameters
        ----------
        timeout: Union[float, None]
            Timeout in seconds to perform commands. This value must be greater
            than zero or ``None``. The default is ``None``. Another function
            raises a timeout exception if the timeout value has elapsed before
            an operation has completed. If the timeout is ``None``, functions
            wait until they're finished, and no timeout exception is raised.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        ValueError
            Raised when the timeout value is less than or equal to 0.
        TypeError
            Raised when the timeout is not a Union[float, None].
        """
        self.__osl_server.set_timeout(timeout)

    def shutdown(self, force: bool = False) -> None:
        """Shut down the optiSLang server.

        This method stops the server from listening for incoming connections, discards
        pending requests, and shuts down the server. In batch mode, the project runs
        until execution finishes and then optiSLang is shut down.

        Parameters
        ----------
        force : bool, optional
            Whether to forcibly shut down a local optiSLang server. The default is
            ``False``. This parameter has no effect when the connection established
            is to a remote optiSLang server. In all cases, an attempt is made to
            shut down the optiSLang server in the proper way. However, if the
            ``force`` parameter is ``True``, after a while, the process is forcibly
            shut down without an exception being raised.

        Raises
        ------
        OslCommunicationError
            Raised when the ``force`` parameter is ``False`` and an error occurs
            while communicating with the server.
        OslCommandError
            Raised when the ``force`` parameter is ``False`` and the command or query fails.
        TimeoutError
            Raised when the ``force`` parameter is ``False`` and the timeout float value expires.
        """
        self.__osl_server.shutdown(force)

    def start(self, wait_for_started: bool = True, wait_for_finished: bool = True) -> None:
        """Start project execution.

        Parameters
        ----------
        wait_for_started : bool, optional
            Whether this function call should wait for optiSLang to start
            the command execution. The default is ``True``, in which case the
            function call does not continue on to the next line of the Python
            script after the command is successfully sent to optiSLang but
            rater waits for execution of the flow inside optiSLang to start.
        wait_for_finished : bool, optional
            Whether this function call should wait for optiSLang to finish
            the command execution. The default is ``True``, in which case the
            function call does not continue on to the next line of the Python
            script after the command is successfully sent to optiSLang but
            rather waits for execution of the flow inside optiSLang to finish.
            This implicitly interprets ``wait_for_started`` as True.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        self.__osl_server.start(wait_for_started, wait_for_finished)

    def stop(self, wait_for_finished: bool = True) -> None:
        """Stop project execution.

        Parameters
        ----------
        wait_for_finished : bool, optional
            Whether this function call should wait for optiSLang to finish
            the command execution. The default is ``True``, in which case the
            function call does not continue on to the next line of the Python
            script after the command is successfully sent to optiSLang but
            rather waits for execution of the flow inside optiSLang to finish.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        self.__osl_server.stop(wait_for_finished)

    def get_osl_server(self) -> Union[OslServer, None]:
        """Get the currently used instance of the OslServer.

        This instance can be used to directly communicate with optiSLang using
        the optiSLang server API.
        """
        return self.__osl_server

    # stop_gently method doesn't work properly in optiSLang 2023 R1. Thus, it was commented out
    # def stop_gently(self, wait_for_finished: bool = True) -> None:
    #     """Stop project execution after the design is finished.

    #     Parameters
    #     ----------
    #     wait_for_finished : bool, optional
    #         Whether this function call should wait for optiSLang to finish
    #         the command execution. The default is ``True``, in which case the
    #         function call does not continue on to the next line of the Python
    #         script after the command is successfully sent to optiSLang but
    #         rather waits for execution of the flow inside optiSLang to finish.

    #     Raises
    #     ------
    #     OslCommunicationError
    #         Raised when an error occurs while communicating with the server.
    #     OslCommandError
    #         Raised when a command or query fails.
    #     TimeoutError
    #         Raised when the timeout float value expires.
    #     """
    #     self.__osl_server.stop_gently(wait_for_finished)



"""Osl logging module."""
from __future__ import annotations

from copy import copy
from datetime import datetime
import logging
from typing import TYPE_CHECKING
import weakref

if TYPE_CHECKING:
    from ansys.optislang.core import Optislang

## Default configuration
LOG_LEVEL = "DEBUG"
FILE_NAME = "pyOptislang.log"

## Formatting

STDOUT_MSG_FORMAT = "%(levelname)s -  %(module)s - %(name)s -  %(funcName)s - %(message)s"
FILE_MSG_FORMAT = STDOUT_MSG_FORMAT

DEFAULT_STDOUT_HEADER = """
LEVEL - MODULE - CLASS - FUNCTION - MESSAGE
"""
DEFAULT_FILE_HEADER = DEFAULT_STDOUT_HEADER
NEW_SESSION_HEADER = f"""
===============================================================================
       NEW SESSION - {datetime.now().strftime("%m/%d/%Y, %H:%M:%S")}
===============================================================================
"""


class OslCustomAdapter(logging.LoggerAdapter):
    """Provides a customized logger with extra inputs."""

    file_handler = None
    std_out_handler = None
    log_level = LOG_LEVEL

    def __init__(
        self,
        logger: logging.Logger,
        extra=None,
    ) -> None:
        """
        Initialize the customized logger with extra inputs.

        Parameters
        ----------
        logger : str, optional
            Level of logging. The default is ``LOG_LEVEL``.
        osl_instance : bool, optional
            Whether to record logs to an output file. The default is ``True``.
        """
        self.logger = logger
        if extra:
            self.extra = weakref.proxy(extra)
        else:
            self.extra = None

        self.file_handler = logger.file_handler
        self.std_out_handler = logger.std_out_handler

    def process(self, msg, kwargs):
        """Modify the message and/or keyword arguments passed to a logging call."""
        kwargs["extra"] = {}
        # These are the extra parameters sent to the log.
        # Here, self.extra is the argument passed to the log records.
        try:
            kwargs["extra"]["instance_name"] = self.extra.name
        except ReferenceError:
            kwargs["extra"]["instance_name"] = "Undefined"
        return msg, kwargs


class OslLogger:
    """Provides the logger class created for each session."""

    file_handler = None
    std_out_handler = None
    log_level = LOG_LEVEL
    instances = {}

    def __init__(
        self,
        loglevel: str = LOG_LEVEL,
        log_to_file: bool = False,
        logfile_name: str = FILE_NAME,
        log_to_stdout: bool = True,
    ) -> None:
        """Initialize the customized logger.

        Parameters
        ----------
        loglevel : str, optional
            Level of logging. The default is ``LOG_LEVEL``.
        log_to_file : bool, optional
            Whether to record logs to an output file. The default is ``False``.
        logfile_name: str, optional
            Name of the output file. The default is ``FILE_NAME``.
        log_to_stdout : bool, optional
            Whether to log to the command line. The default is ``False``.
        """
        # Create logger
        self.logger = logging.getLogger("pyoptislang_global")
        self.set_log_level(loglevel)
        self.logger.propagate = True

        if log_to_file:
            self.add_file_handler(logfile_name=logfile_name, loglevel=loglevel)

        if log_to_stdout:
            self.add_std_out_handler(loglevel=loglevel)

    def set_log_level(
        self,
        loglevel: str,
    ) -> None:
        """Set the log level of the object and its handlers.

        Parameters
        ----------
        loglevel : str
            Level of logging. The default is ``LOG_LEVEL``.
        """
        self.logger.setLevel(loglevel)
        for handler in self.logger.handlers:
            handler.setLevel(loglevel)
        self.log_level = loglevel

    def add_file_handler(
        self,
        logfile_name: str = FILE_NAME,
        loglevel: str = LOG_LEVEL,
    ) -> None:
        """Add a file handler (output file) to the logger.

        Parameters
        ----------
        logger: logging.Logger
            Logger for creating the output file.
        logfile_name: str, optional
            Name of the output file. The default is ``FILE_NAME``.
        loglevel : str, optional
            Level of logging. The default is ``LOG_LEVEL``.
        """
        file_handler = logging.FileHandler(logfile_name)
        file_handler.setFormatter(logging.Formatter(FILE_MSG_FORMAT))
        file_handler.setLevel(loglevel)
        self.logger.addHandler(file_handler)
        self.file_handler = file_handler
        file_handler.stream.write(NEW_SESSION_HEADER)
        file_handler.stream.write(DEFAULT_FILE_HEADER)

    def add_std_out_handler(
        self,
        loglevel: str = LOG_LEVEL,
    ) -> None:
        """Add standard output to the terminal.

        Parameters
        ----------
        loglevel : str, optional
            Level of logging. The default is ``LOG_LEVEL``.
        """
        std_out_handler = logging.StreamHandler()
        std_out_handler.setLevel(loglevel)
        std_out_handler.setFormatter(logging.Formatter(FILE_MSG_FORMAT))
        self.logger.addHandler(std_out_handler)
        self.std_out_handler = std_out_handler
        std_out_handler.stream.write(NEW_SESSION_HEADER)
        std_out_handler.stream.write(DEFAULT_STDOUT_HEADER)

    def create_logger(self, new_logger_name: str, level: str = None) -> logging.Logger:
        """Create a logger for the Optislang instance.

        Parameters
        ----------
        new_logger_name : str
            Name of the logger.
        level: str, opt
            Level of logging. The default is ``None``, in which gase the global log level is set.

        Returns
        -------
        logging.logger
            Logger class.
        """
        new_logger = logging.getLogger(new_logger_name)
        new_logger.std_out_handler = None
        new_logger.file_handler = None

        if level is None:
            level = self.log_level
        level = level.upper()
        new_logger.setLevel(level)

        if self.file_handler:
            new_logger.file_handler = copy(self.file_handler)
            new_logger.addHandler(new_logger.file_handler)
            new_logger.file_handler.setLevel(level)

        if self.std_out_handler:
            new_logger.std_out_handler = copy(self.std_out_handler)
            new_logger.addHandler(new_logger.std_out_handler)
            new_logger.std_out_handler.setLevel(level)

        new_logger.propagate = True
        return new_logger

    def add_child_logger(
        self,
        child_logger_name: str,
    ) -> None:
        """Call the ``create_logger`` method to add a child logger.

        This looger is more general than the main logger.

        Parameters
        ----------
        child_logger_name : str
            Name of the  child logger.
        """
        if type(child_logger_name) is not str:
            raise ValueError("Expected input child_logger_name: str")
        child_logger = self.logger.name + "." + child_logger_name
        self.instances[child_logger] = self.create_logger(child_logger)

    def add_instance_logger(
        self, instance_name: str, osl_instance: Optislang, level: str = None
    ) -> OslCustomAdapter:
        """Add a logger for an Optislang instance.

        Parameters
        ----------
        instance_name : str
            Name of the instance logger.
        osl_instance: Optislang
            Optislang instance object. This object should contain the ``name``
            attribute.
        level: str, opt
            Level of logging.

        Returns
        -------
        OslCustomAdapter
            Customized logger with extra inputs.
        """
        if type(instance_name) is not str:
            raise ValueError("Expected input instance_name: str")

        # Rename instance if already exists
        counter = 0
        name = instance_name
        while instance_name in logging.root.manager.__dict__.keys():
            counter += 1
            instance_name = name + str(counter)

        self.instances[instance_name] = OslCustomAdapter(
            self.create_logger(instance_name, level), osl_instance
        )

        return self.instances[instance_name]

    def __getitem__(self, instance_name: str) -> logging.Logger:
        """Get a logger by an instance name.

        Parameters
        ----------
        instance_name : str
            Name of the instance.
        Returns
        -------
        self.instances[instance_name]: logging.Logger()
            Logger class.
        """
        if instance_name in self.instances.keys():
            return self.instances[instance_name]
        else:
            raise KeyError(f"There is no instances with name {instance_name}")



"""Create workflow."""
###############################################################################
# (c) 2021 ANSYS, Inc. Unauthorized use, distribution, or duplication is prohibited.

# create robustness flow for oscillator python example

# in batch:
#      optislang --new oscillator_py.opf --python oscillator_robustness_arsm.py --run --batch

# in optiSlang:
#      open optiSLang and use 'run script file...' in optiSlang python console

# find a description of the oscillator example in optiSLang tutorial section
###############################################################################

###############################################################################
# Import needed modules
###############################################################################
import os

from dynardo_py_algorithms import DOETYPES
import py_doe_settings
from py_os_criterion import DesignStatus, PyOSCriterion, PyOSCriterionContainer

###############################################################################
# Create ARSM, MOP, Robustness and PostProcessing
###############################################################################
arsm = actors.ARSMActor("ARSM")
arsm.auto_save_mode = AS_ACTOR_FINISHED
add_actor(arsm)

robust = actors.RobustnessActor("Robustness")
robust.auto_save_mode = AS_ACTOR_FINISHED
add_actor(robust)

post_proc_opt = actors.PostprocessingActor("PostProcessing_Optimization")
add_actor(post_proc_opt)

post_proc_robust = actors.PostprocessingActor("PostProcessing_Robustness")
add_actor(post_proc_robust)

mop = actors.MOPActor("MOP")
mop.show_pp_when_avaiable = True
add_actor(mop)

connect(robust, "OMDBPath", post_proc_robust, "IMDBPath")
connect(post_proc_robust, "OReducedMDBPath", mop, "IMDBPath")
connect(robust, "OParameterManager", mop, "IParameterManager")
connect(arsm, "OBestDesigns", robust, "INominalDesigns")
connect(arsm, "OMDBPath", post_proc_opt, "IMDBPath")

###############################################################################
# Create oscillator example @ example_system
###############################################################################
# get example path "OSL_EXAMPLES": .\pyoptislang\src\ansys\optislang\core\examples
example_system = arsm
path = os.path.join(
    os.environ["OSL_EXAMPLES"], "00_run_script", "oscillator", "scripts", "_create_oscillator.py"
)
exec(open(path).read())
arsm = example_system

example_system = robust
exec(open(path).read())
robust = example_system

###############################################################################
# Setup ARSM
###############################################################################
settings = arsm.doe_settings
settings.sampling_method = DOETYPES.KOSHALLINEAR
arsm.doe_settings = settings

###############################################################################
# Setup Criteria
###############################################################################
criteria_exprs = PyOSCriterionContainer()
criteria_exprs.add("Objective", PyOSCriterion(DesignStatus.MIN, "x_max"))
criteria_exprs.add("Constraint", PyOSCriterion(DesignStatus.LESSEQUAL, "omega_damped", "8.0"))
arsm.criteria = criteria_exprs
robust.criteria = criteria_exprs

###############################################################################
# Setup Postprocessing
###############################################################################
post_proc_opt.pp_mode = actors.PP3_AUTOMATIC
post_proc_opt.show_pp_during_run = False
post_proc_opt.show_reduced_pp_when_avaiable = False
post_proc_opt.wait_for_pp_to_finish = True
post_proc_opt.force_classic = False

post_proc_robust.pp_mode = actors.PP3_AUTOMATIC
post_proc_robust.show_pp_during_run = False
post_proc_robust.show_reduced_pp_when_avaiable = False
post_proc_robust.wait_for_pp_to_finish = True
post_proc_robust.force_classic = False



"""Download files from repository."""

import os
from pathlib import Path
from typing import Tuple

import ansys.optislang.core.examples as examples

# TODO: implement automatic download from online repository
# EXAMPLE_REPO = "https://github.com/ansys/pyoptislang/tree/main/examples/files"


def _download_files(scriptname: str) -> Tuple[Path, ...]:
    """Check if files exists, otherwise download them. Return path to files then."""
    # check if file was downloaded
    for file_path in examples.example_files[scriptname]:
        if not file_path.is_file():
            NotImplementedError("Automatic download not implemented.")

    return examples.example_files[scriptname]


def _download_script(scriptname: str) -> Path:
    """Check if file exists, otherwise download it. Return path to file then."""
    # check if script was downloaded, download it if not
    local_path = examples.example_scripts[scriptname]

    if not local_path.is_file():
        NotImplementedError("Automatic download not implemented.")

    return local_path


def get_files(scriptname: str) -> Tuple[Path, Tuple[Path, ...]]:
    """Get tuple of files necessary for running example.

    Parameters
    ----------
    scriptname: str
        Name of the example script (without ``.py``).

    Returns
    -------
    Tuple[Path, Tuple[Path, ...]]
        Tuple[0]: path to script
        Tuple[1]: tuple of paths to files necessary for running script
    """
    if examples.example_files[scriptname] is not None:
        file_path = _download_files(scriptname)
    else:
        file_path = None

    if examples.example_scripts[scriptname] is not None:
        script_path = _download_script(scriptname)
    else:
        script_path = None

    return (script_path, file_path)



"""Module for generation of all server commands."""
import json
from typing import Dict, Iterable, Sequence, Union

_APPLY_WIZARD = "APPLY_WIZARD"
_CLOSE = "CLOSE"
_CONNECT_NODES = "CONNECT_NODES"
_CREATE_INPUT_SLOT = "CREATE_INPUT_SLOT"
_CREATE_NODE = "CREATE_NODE"
_CREATE_OUTPUT_SLOT = "CREATE_OUTPUT_SLOT"
_CREATE_START_DESIGNS = "CREATE_START_DESIGNS"
_DISCONNECT_SLOT = "DISCONNECT_SLOT"
_EVALUATE_DESIGN = "EVALUATE_DESIGN"
_EXPORT_DESIGNS = "EXPORT_DESIGNS"
_FINALIZE = "FINALIZE"
_LINK_REGISTERED_FILE = "LINK_REGISTERED_FILE"
_NEW = "NEW"
_OPEN = "OPEN"
_PAUSE = "PAUSE"
_REEVALUATE_STATE = "REEVALUATE_STATE"
_REFRESH_LISTENER_REGISTRATION = "REFRESH_LISTENER_REGISTRATION"
_REGISTER_FILE = "REGISTER_FILE"
_REGISTER_LISTENER = "REGISTER_LISTENER"
_REGISTER_LOCATIONS_AS_PARAMETER = "REGISTER_LOCATIONS_AS_PARAMETER"
_REGISTER_LOCATIONS_AS_RESPONSE = "REGISTER_LOCATIONS_AS_RESPONSE"
_REMOVE_NODE = "REMOVE_NODE"
_RE_REGISTER_LOCATIONS_AS_PARAMETER = "RE_REGISTER_LOCATIONS_AS_PARAMETER"
_RE_REGISTER_LOCATIONS_AS_RESPONSE = "RE_REGISTER_LOCATIONS_AS_RESPONSE"
_RESET = "RESET"
_RESTART = "RESTART"
_RESUME = "RESUME"
_RUN_PYTHON_SCRIPT = "RUN_PYTHON_SCRIPT"
_RUN_REGISTERED_FILES_ACTIONS = "RUN_REGISTERED_FILES_ACTIONS"
_SAVE = "SAVE"
_SAVE_AS = "SAVE_AS"
_SAVE_COPY = "SAVE_COPY"
_SET_ACTOR_PROPERTY = "SET_ACTOR_PROPERTY"
_SET_ACTOR_SETTING = "SET_ACTOR_SETTING"
_SET_ACTOR_STATE_PROPERTY = "SET_ACTOR_STATE_PROPERTY"
_SET_PLACEHOLDER_VALUE = "SET_PLACEHOLDER_VALUE"
_SET_PROJECT_SETTING = "SET_PROJECT_SETTING"
_SET_REGISTERED_FILE_VALUE = "SET_REGISTERED_FILE_VALUE"
_SET_START_DESIGNS = "SET_START_DESIGNS"
_SET_SUCCEEDED_STATE = "SET_SUCCEEDED_STATE"
_SHOW_DIALOG = "SHOW_DIALOG"
_SHOW_NODE_DIALOG = "SHOW_NODE_DIALOG"
_SHUTDOWN = "SHUTDOWN"
_SHUTDOWN_WHEN_FINISHED = "SHUTDOWN_WHEN_FINISHED"
_START = "START"
_STOP = "STOP"
_STOP_GENTLY = "STOP_GENTLY"
_SUBSCRIBE_FOR_PUSH_NOTIFICATIONS = "SUBSCRIBE_FOR_PUSH_NOTIFICATIONS"
_UNLINK_REGISTERED_FILE = "UNLINK_REGISTERED_FILE"
_UNREGISTER_FILE = "UNREGISTER_FILE"
_UNREGISTER_LISTENER = "UNREGISTER_LISTENER"
_WRITE_MONITORING_DATABASE = "WRITE_MONITORING_DATABASE"
_builtin = "builtin"


def apply_wizard(
    actor_uid: str,
    type_: str,
    use_existing_system: bool = None,
    usage_mode: str = None,
    parent_hwnd: str = None,
    password: str = None,
) -> str:
    """Generate JSON string of apply_wizzard command.

    Parameters
    ----------
    actor_uid: str
        Unique identifying actor of the object.
    type_: str
        Node or system type, supported values are:
        ["solver", "sensitivity", "optimization", "robustness", "reevaluation"].
    use_existing_system: bool, opt
        True or false.
    usage_mode: str, opt
        Usage mode, e.g. "EXPERT".
    parent_hwnd: str, opt
        Parent hwnd.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of apply_wizzard command.
    """
    args = {}
    supported_values = ["solver", "sensitivity", "optimization", "robustness", "reevaluation"]
    if type_ in supported_values:
        args["type"] = type_
    else:
        raise TypeError(
            f"Unsuppored value of ``type_``: {type_}, supported values are: {supported_values}"
        )
    if use_existing_system is not None:
        args["use_existing_system"] = use_existing_system
    if usage_mode is not None:
        args["usage_mode"] = usage_mode
    # TODO: and/or:  The request must contain the actor_uid entry as well as the args entry,
    # containing the argument type and optionally the arguments usage_mode,
    # use_existing_system and/or parent_hwnd."
    if parent_hwnd is not None:
        args["parent_hwnd"] = parent_hwnd

    return _to_json(
        _gen_server_command(
            command=_APPLY_WIZARD, actor_uid=actor_uid, args=args, password=password
        )
    )


def close(password: str = None) -> str:
    """Generate JSON string of close command.

    Parameters
    ----------
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of close command.
    """
    return _to_json(_gen_server_command(command=_CLOSE, password=password))


def connect_nodes(
    from_actor_uid: str, from_slot: str, to_actor_uid: str, to_slot: str, password: str = None
) -> str:
    """Generate JSON string of connect_nodes command.

    Parameters
    ----------
    from_actor_uid: str
        Uid of connection source.
    from_slot: str
        Slot of connection source.
    to_actor_uid: str
        Uid of connection target.
    to_slot: str
        Slot of connection target.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of connect_nodes command.
    """
    args = {}
    args["from_actor_uid"] = from_actor_uid
    args["from_slot"] = from_slot
    args["to_actor_uid"] = to_actor_uid
    args["to_slot"] = to_slot

    return _to_json(_gen_server_command(command=_CONNECT_NODES, args=args, password=password))


def create_input_slot(
    actor_uid: str, slot_name: str, type_hint: str = None, password: str = None
) -> str:
    """Generate JSON string of create_input_slot command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    slot_name: str
        Name of slot.
    type_hint: str, opt
        Type of hint.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of create_input_slot command.
    """
    args = {}
    args["slot_name"] = slot_name
    if type_hint is not None:
        args["type_hint"] = type_hint

    return _to_json(
        _gen_server_command(
            command=_CREATE_INPUT_SLOT, actor_uid=actor_uid, args=args, password=password
        )
    )


def create_node(
    type_: str, name: str, parent_uid: str = None, design_flow: str = None, password: str = None
) -> str:
    """Generate JSON string of create_node command.

    Parameters
    ----------
    type_: str
        Type.
    name: str
        Name.
    parent_uid: str, opt
        Parent uid entry.
    design_flow: str, opt
        Design flow, optional values are ["RECEIVE", "SEND", "RECEIVE_SEND"].
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of create_node command.
    """
    args = {}
    args["type"] = type_
    args["name"] = name
    if parent_uid is not None:
        args["parent_uid"] = parent_uid
    if design_flow not in [None, "NONE"]:
        args["design_flow"] = design_flow

    return _to_json(_gen_server_command(command=_CREATE_NODE, args=args, password=password))


def create_output_slot(
    actor_uid: str, slot_name: str, type_hint: str = None, password: str = None
) -> str:
    """Generate JSON string of create_output_slot command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    slot_name: str
        Name of the slot.
    type_hint: str
        Type of the hint.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of create_output_slot command.
    """
    args = {}
    args["slot_name"] = slot_name
    if type_hint is not None:
        args["type_hint"] = type_hint
    return _to_json(
        _gen_server_command(
            command=_CREATE_OUTPUT_SLOT, actor_uid=actor_uid, args=args, password=password
        )
    )


def create_start_designs(
    actor_uid: str,
    sampling_type: str,
    number_of_levels: int = None,
    number_of_samples: int = None,
    password: str = None,
) -> str:
    """Generate JSON string of create_start_design command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    sampling_type: str
        Sampling type, e.g. "fullfactorial".
    number_of_levels: int, opt
        Number of levels.
    number_of_samples: int, opt
        Number of samples.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of create_start_design command.
    """
    args = {}
    args["sampling_type"] = sampling_type
    if (number_of_levels is not None) and (number_of_samples is not None):
        raise TypeError(f"Please specify either ``number_of_levels`` or ``number of samples``.")
    elif number_of_levels is not None:
        args["number_of_levels"] = number_of_levels
    elif number_of_samples is not None:
        args["number_of_samples"] = number_of_samples

    return _to_json(
        _gen_server_command(
            command=_CREATE_START_DESIGNS, actor_uid=actor_uid, args=args, password=password
        )
    )


def disconnect_slot(actor_uid: str, slot_name: str, direction: str, password: str = None) -> str:
    """Generate JSON string of disconnect_slot command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    slot_name: str
        Name of the slot.
    direction: str
        Direction.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of disconnect_slot command.
    """
    args = {}
    args["slot_name"] = slot_name
    args["direction"] = direction
    return _to_json(
        _gen_server_command(
            command=_DISCONNECT_SLOT, actor_uid=actor_uid, args=args, password=password
        )
    )


def evaluate_design(parameters: Dict, password: str = None) -> str:
    """Generate JSON string of evaluate_design command.

    Parameters
    ----------
    parameters: Dict
        Dictionary of parameters.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of evaluate_design command.
    """
    args = {}
    args["parameters"] = parameters
    return _to_json(_gen_server_command(command=_EVALUATE_DESIGN, args=args, password=password))


def export_designs(
    actor_uid: str, path: str, format: str = None, csv_separator: str = None, password: str = None
) -> str:
    """Generate JSON string of export_designs command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    path: str
       Path.
    format: str, opt
        Type_hint format.
    csv_separator: str, opt
        CSV separator.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of export_designs command.
    """
    args = {}
    args["path"] = path
    if format is not None:
        args["format"] = format
    if csv_separator is not None:
        args["csv_separator"] = csv_separator

    return _to_json(
        _gen_server_command(
            command=_EXPORT_DESIGNS, actor_uid=actor_uid, args=args, password=password
        )
    )


def finalize(actor_uid: str, password: str = None) -> str:
    """Generate JSON string of finalize command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of finalize command.
    """
    return _to_json(_gen_server_command(command=_FINALIZE, actor_uid=actor_uid, password=password))


def link_registered_file(actor_uid: str, uid: str, password: str = None) -> str:
    """Generate JSON string of link_registered_file command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    uid: str
        Uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of link_registered_file command.
    """
    args = {}
    args["uid"] = uid
    return _to_json(
        _gen_server_command(
            command=_LINK_REGISTERED_FILE, actor_uid=actor_uid, args=args, password=password
        )
    )


def new(password: str = None) -> str:
    """Generate JSON string of ``new`` command.

    Parameters
    ----------
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``new`` command.
    """
    return _to_json(_gen_server_command(command=_NEW, password=password))


def open(path: str, do_force: bool, do_restore: bool, do_reset: bool, password: str = None) -> str:
    """Generate JSON string of ``open`` command.

    Parameters
    ----------
    path: str
        path.
    do_force: bool
        True/False.
    do_restore: bool
        True/False.
    do_reset: bool
        True/False.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``open`` command.
    """
    args = {}
    args["path"] = path
    args["do_force"] = do_force
    args["do_restore"] = do_restore
    args["do_reset"] = do_reset

    return _to_json(_gen_server_command(command=_OPEN, args=args, password=password))


def pause(password: str = None) -> str:
    """Generate JSON string of pause command.

    Parameters
    ----------
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of pause command.
    """
    return _to_json(_gen_server_command(command=_PAUSE, password=password))


def reevaluate_state(actor_uid: str, hid: str = None, password: str = None) -> str:
    """Generate JSON string of reevaluate_state command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    hid: str
        Hid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of reevaluate_state command.
    """
    return _to_json(
        _gen_server_command(
            command=_REEVALUATE_STATE, actor_uid=actor_uid, hid=hid, password=password
        )
    )


def refresh_listener_registration(uid: str, password: str = None) -> str:
    """Generate JSON string of refresh_listener_registration command.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of refresh_listener_registration command.
    """
    args = {}
    args["uid"] = uid
    return _to_json(
        _gen_server_command(command=_REFRESH_LISTENER_REGISTRATION, args=args, password=password)
    )


def register_file(
    uid: str = None,
    ident: str = None,
    local_location: Dict = None,
    action: str = None,
    password: str = None,
) -> str:
    """Generate JSON string of register_file command.

    Parameters
    ----------
    uid: str, opt
        Uid entry.
    ident: str, opt
        Ident.
    local_location: Dict, opt
        Dictionary specifying location, e. g. { "split_path": {"head": "",
        "tail": "C:/samples_path/result.txt" }, "base_path": "",
        "base_path_mode": "ABSOLUTE_PATH" }.
    action: str, opt
        Action, e. g. "Send"
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of register_file command.
    """
    # TODO: "and so on: specify all possibilities"
    args = {}
    if uid is not None:
        args["uid"] = uid
    if ident is not None:
        args["ident"] = ident
    if local_location is not None:
        args["local_location"] = local_location
    if action is not None:
        args["action"] = action

    return _to_json(_gen_server_command(command=_REGISTER_FILE, args=args, password=password))


def register_listener(
    id: str = None,
    host: str = None,
    port: int = None,
    timeout: int = None,
    notifications: Iterable[str] = None,
    password: str = None,
    listener_uid: str = None,
) -> str:
    """Generate JSON string of register_listener command.

    Parameters
    ----------
    id: str, opt
        Id of the local listener.
    host: str, opt
        IP of the TCP listener
    port: int, opt
        Port of the TCP listener.
    time_out: int, opt
        Unregister policy timeout in ms, default 60000 ms.
    notifications: Sequence, opt
        Notifications.
    password : str, opt
        Password.
    listener_uid : str, opt
        Listener UID.

    Returns
    -------
    str
        JSON string of register_listener command.
    """
    args = {}
    if (id is not None) and ((host is not None) or (port is not None)):
        raise TypeError(f"Please specify either ``id`` or (``host`` and ``port``).")
    elif (id is None) and ((host is None) or (port is None)):
        raise TypeError(f"Please specify either ``id`` or (``host`` and ``port``).")
    elif id is not None:
        args["id"] = id
    else:
        args["host"] = host
        args["port"] = port

    if timeout is not None:
        args["timeout"] = timeout
    if notifications is not None:
        args["notifications"] = notifications
    if listener_uid is not None:
        args["uid"] = listener_uid

    return _to_json(_gen_server_command(command=_REGISTER_LISTENER, args=args, password=password))


def register_locations_as_parameter(actor_uid: str, password: str = None) -> str:
    """Generate JSON string of ``register_location_as_parameter`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``register_location_as_parameter`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_REGISTER_LOCATIONS_AS_PARAMETER, actor_uid=actor_uid, password=password
        )
    )


def register_locations_as_response(actor_uid: str, password: str = None) -> str:
    """Generate JSON string of ``register_location_as_response`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``register_location_as_response`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_REGISTER_LOCATIONS_AS_RESPONSE, actor_uid=actor_uid, password=password
        )
    )


def remove_node(actor_uid: str, password: str = None) -> str:
    """Generate JSON string of ``remove node`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``remove_node`` command.
    """
    return _to_json(
        _gen_server_command(command=_REMOVE_NODE, actor_uid=actor_uid, password=password)
    )


def re_register_locations_as_parameter(actor_uid: str, password: str = None) -> str:
    """Generate JSON string of ``re_register_locations_as_parameter`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``re_register_locations_as_parameter`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_RE_REGISTER_LOCATIONS_AS_PARAMETER, actor_uid=actor_uid, password=password
        )
    )


def re_register_locations_as_response(actor_uid: str, password: str = None) -> str:
    """Generate JSON string of ``re_register_locations_as_response`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``re_register_locations_as_response`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_RE_REGISTER_LOCATIONS_AS_RESPONSE, actor_uid=actor_uid, password=password
        )
    )


def reset(actor_uid: str = None, hid: str = None, password: str = None) -> str:
    """Generate JSON string of ``reset`` command.

    Parameters
    ----------
    actor_uid: str, opt
        Actor uid entry. A Hirearchical ID (hid) is required.
    hid: str, opt
        Hid entry. The actor uid is required.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``reset`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hirearchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")
    return _to_json(
        _gen_server_command(command=_RESET, actor_uid=actor_uid, hid=hid, password=password)
    )


def restart(actor_uid: str = None, hid: str = None, password: str = None) -> str:
    """Generate JSON string of ``restart`` command.

    Parameters
    ----------
    actor_uid: str, opt
        Actor uid entry. A Hirearchical ID (hid) is required.
    hid: str, opt
        Hid entry. The actor uid is required.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``restart`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hirearchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")
    return _to_json(
        _gen_server_command(command=_RESTART, actor_uid=actor_uid, hid=hid, password=password)
    )


def resume(password: str = None) -> str:
    """Generate JSON string of ``resume`` command.

    Parameters
    ----------
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``resume`` command.
    """
    return _to_json(_gen_server_command(command=_RESUME, password=password))


def run_python_script(
    script: str,
    args_: list = None,
    password: str = None,
) -> str:
    """Generate JSON string of register_listener command.

    Parameters
    ----------
    script: str
        Path of the script.
    args: list, opt
        IP of the TCP listener.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of register_listener command.
    """
    args = {}
    args["script"] = script
    if args_ is not None:
        args["args"] = args_

    return _to_json(_gen_server_command(command=_RUN_PYTHON_SCRIPT, args=args, password=password))


def run_registered_files_actions(
    uid: str = None,
    password: str = None,
) -> str:
    """Generate JSON string of ``run registered files actions`` command.

    Parameters
    ----------
    uid: str, opt
        Uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``run registered files actions`` command.
    """
    args = None
    if uid is not None:
        args = {}
        args["uid"] = uid

    return _to_json(
        _gen_server_command(command=_RUN_REGISTERED_FILES_ACTIONS, args=args, password=password)
    )


def save(password: str = None) -> str:
    """Generate JSON string of ``save`` command.

    Parameters
    ----------
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``save`` command.
    """
    return _to_json(_gen_server_command(command=_SAVE, password=password))


def save_as(
    path: str, do_force: bool, do_restore: bool, do_reset: bool, password: str = None
) -> str:
    """Generate JSON string of ``save_as`` command.

    Parameters
    ----------
    path: str
        path.
    do_force: bool
        True/False.
    do_restore: bool
        True/False.
    do_reset: bool
        True/False.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``save_as`` command.
    """
    args = {}
    args["path"] = path
    args["do_force"] = do_force
    args["do_restore"] = do_restore
    args["do_reset"] = do_reset

    return _to_json(_gen_server_command(command=_SAVE_AS, args=args, password=password))


def save_copy(path: str, password: str = None) -> str:
    """Generate JSON string of ``save_copy`` command.

    Parameters
    ----------
    path: str
        path.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``save_copy`` command.
    """
    args = {}
    args["path"] = path

    return _to_json(_gen_server_command(command=_SAVE_COPY, args=args, password=password))


def set_actor_property(actor_uid: str, name: str, value: str, password: str = None) -> str:
    """Generate JSON string of ``set_actor_property`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    name: str
        Property name.
    value: str
        Value.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``set_actor_property`` command.
    """
    args = {}
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(
            command=_SET_ACTOR_PROPERTY, actor_uid=actor_uid, args=args, password=password
        )
    )


def set_actor_setting(actor_uid: str, name: str, value: str, password: str = None) -> str:
    """Generate JSON string of ``set actor setting`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    name: str
        Property name.
    value: str
        Value.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``set actor setting`` command.
    """
    args = {}
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(
            command=_SET_ACTOR_SETTING, actor_uid=actor_uid, args=args, password=password
        )
    )


def set_actor_state_property(actor_uid: str, name: str, value: str, password: str = None) -> str:
    """Generate JSON string of ``set actor state property`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    name: str
        Property name.
    value: str
        Value.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``set actor state property`` command.
    """
    args = {}
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(
            command=_SET_ACTOR_STATE_PROPERTY, actor_uid=actor_uid, args=args, password=password
        )
    )


def set_placeholder_value(name: str, value: str, password: str = None) -> str:
    """Generate JSON string of ``set placeholder value`` command.

    Parameters
    ----------
    name: str
        Property name.
    value: str
        Value.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``set placeholder value`` command.
    """
    args = {}
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(command=_SET_PLACEHOLDER_VALUE, args=args, password=password)
    )


def set_project_setting(name: str, value: str, password: str = None) -> str:
    """Generate JSON string of ``set project settings`` command.

    Parameters
    ----------
    name: str
        Property name.
    value: str
        Value.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``set project settings`` command.
    """
    args = {}
    args["name"] = name
    args["value"] = value

    return _to_json(_gen_server_command(command=_SET_PROJECT_SETTING, args=args, password=password))


def set_registered_file_value(
    uid: str, name: str, value: Union[str, Dict], password: str = None
) -> str:
    """Generate JSON string of ``set registered file value`` command.

    Parameters
    ----------
    uid: str
        Uid entry.
    name: str
        Property name.
    value: Union[str, Dict]
        Value.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``set registered file value`` command.
    """
    args = {}
    args["uid"] = uid
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(command=_SET_REGISTERED_FILE_VALUE, args=args, password=password)
    )


def set_start_designs(actor_uid: str, start_designs: Dict, password: str = None) -> str:
    """Generate JSON string of ``set start designs`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    start_designs: Dict
        Dictionary of settings.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``set start designs`` command.
    """
    # TODO: function to create start_settings?
    args = {}
    args["start_designs"] = start_designs

    return _to_json(
        _gen_server_command(
            command=_SET_START_DESIGNS, actor_uid=actor_uid, args=args, password=password
        )
    )


def set_succeeded_state(actor_uid: str, hid: str = None, password: str = None) -> str:
    """Generate JSON string of ``set succeeded state`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    hid: str, opt
        Hid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``set succeeded state`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_SET_SUCCEEDED_STATE, actor_uid=actor_uid, hid=hid, password=password
        )
    )


def show_dialog(
    type_: str, usage_mode: str = None, parent_hwnd: str = None, password: str = None
) -> str:
    """Generate JSON string of ``show dialog`` command.

    Parameters
    ----------
    type_: str
        Argument type,
        for application dialogs: [ "about", "help", "settings", "plugin" ]
        for project dialogs: [ "project_settings", "project_overview", "license_management",
        "registered_files", "purge", "load_from", "save_to" ].
    usage_mode: str, opt
        Usage mode. e.g. "EXPERT".
    parent_hwnd: str, opt
        Parent hwnd.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``show dialog`` command.
    """
    application_dialogs = ["about", "help", "settings", "plugin"]
    project_dialogs = [
        "project_settings",
        "project_overview",
        "license_management",
        "registered_files",
        "purge",
        "load_from",
        "save_to",
    ]

    if not ((type_ in application_dialogs) or (type_ in project_dialogs)):
        raise TypeError(
            f"Unsuppored value of ``type_``: {type_}, supported values for application dialogs: "
            "{application_dialogs} and for project dialogs: {project_dialogs}"
        )
    args = {}
    args["type"] = type_
    if usage_mode is not None:
        args["usage_mode"] = usage_mode
    if parent_hwnd is not None:
        args["parent_hwnd"] = parent_hwnd

    return _to_json(_gen_server_command(command=_SHOW_DIALOG, args=args, password=password))


def show_node_dialog(
    actor_uid: str,
    blocking: bool = None,
    type_: str = None,
    usage_mode: str = None,
    parent_hwnd: str = None,
    password: str = None,
) -> str:
    """Generate JSON string of ``show node dialog`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    blocking: bool, opt
        True/False.
    type_: str, opt
        Type, e.g. "help".
    usage_mode: str, opt
        Usage mode. e.g. "EXPERT".
    parent_hwnd: str, opt
        Parent hwnd.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``show node dialog`` command.
    """
    args = None
    if not (
        (blocking is None) and (type_ is None) and (usage_mode is None) and (parent_hwnd is None)
    ):
        args = {}
        if blocking is not None:
            args["blocking"] = blocking
        if type_ is not None:
            args["type"] = type_
        if usage_mode is not None:
            args["usage_mode"] = usage_mode
        if parent_hwnd is not None:
            args["parent_hwnd"] = parent_hwnd

    return _to_json(
        _gen_server_command(
            command=_SHOW_NODE_DIALOG, actor_uid=actor_uid, args=args, password=password
        )
    )


def shutdown(password: str = None) -> str:
    """Generate JSON string of ``shutdown`` command.

    Parameters
    ----------
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``shutdown`` command.
    """
    return _to_json(_gen_server_command(command=_SHUTDOWN, password=password))


def shutdown_when_finished(password: str = None) -> str:
    """Generate JSON string of ``shutdown_when_finished`` command.

    Parameters
    ----------
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``shutdown_when_finished`` command.
    """
    return _to_json(_gen_server_command(command=_SHUTDOWN_WHEN_FINISHED, password=password))


def start(actor_uid: str = None, hid: str = None, password: str = None) -> str:
    """Generate JSON string of ``start`` command.

    Parameters
    ----------
    actor_uid: str, opt
        Actor uid entry. A Hirearchical ID (hid) is required.
    hid: str, opt
        Hid entry. The actor uid is required.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``start`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hirearchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")

    return _to_json(
        _gen_server_command(command=_START, actor_uid=actor_uid, hid=hid, password=password)
    )


def stop(actor_uid: str = None, hid: str = None, password: str = None) -> str:
    """Generate JSON string of ``stop`` command.

    Parameters
    ----------
    actor_uid: str, opt
        Actor uid entry. A Hirearchical ID (hid) is required.
    hid: str, opt
        Hid entry. The actor uid is required.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``stop`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hirearchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")

    return _to_json(
        _gen_server_command(command=_STOP, actor_uid=actor_uid, hid=hid, password=password)
    )


def stop_gently(actor_uid: str = None, hid: str = None, password: str = None) -> str:
    """Generate JSON string of ``stop_gently`` command.

    Parameters
    ----------
    actor_uid: str, opt
        Actor uid entry. A Hirearchical ID (hid) is required.
    hid: str, opt
        Hid entry. The actor uid is required.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``stop_gently`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hirearchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")
    return _to_json(
        _gen_server_command(command=_STOP_GENTLY, actor_uid=actor_uid, hid=hid, password=password)
    )


def subscribe_for_push_notifications(
    uid: str, notifications: Sequence, node_types: Sequence = None, password: str = None
) -> str:
    """Generate JSON string of ``subscribe_for_push_notifications`` command.

    Parameters
    ----------
    uid: str
        Uid entry.
    notifications: Sequence
        Either ["ALL"] or Sequence picked from below options:
            Server: [ "SERVER_UP", "SERVER_DOWN" ] (always be sent by default).
            Logging: [ "LOG_INFO", "LOG_WARNING", "LOG_ERROR", "LOG_DEBUG" ].
            Project: [ "EXECUTION_STARTED", "PROCESSING_STARTED", "EXECUTION_FINISHED",
            "NOTHING_PROCESSED", "CHECK_FAILED", "EXEC_FAILED" ].
            Nodes: [ "ACTOR_STATE_CHANGED", "ACTOR_ACTIVE_CHANGED", "ACTOR_NAME_CHANGED",
            ACTOR_CONTENTS_CHANGED", "ACTOR_DATA_CHANGED" ].
    node_types: Sequence, opt
       Node types, e.g. ["Sensitivity", "AnsysWorkbench"].
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``subscribe_for_push_notifications`` command.
    """
    server = ["SERVER_UP", "SERVER_DOWN"]
    logging = ["LOG_INFO", "LOG_WARNING", "LOG_ERROR", "LOG_DEBUG"]
    project = [
        "EXECUTION_STARTED",
        "PROCESSING_STARTED",
        "EXECUTION_FINISHED",
        "NOTHING_PROCESSED",
        "CHECK_FAILED",
        "EXEC_FAILED",
    ]
    nodes = [
        "ACTOR_STATE_CHANGED",
        "ACTOR_ACTIVE_CHANGED",
        "ACTOR_NAME_CHANGED",
        "ACTOR_CONTENTS_CHANGED",
        "ACTOR_DATA_CHANGED",
    ]
    all_opts = set(server + logging + project + nodes)

    args = {}
    args["uid"] = uid

    if notifications == ["ALL"]:
        args["notifications"] = notifications
    else:
        if not all(item in all_opts for item in notifications):
            invalid_items = list(sorted(all_opts - set(notifications)))
            raise TypeError(
                f"Unsuppored values of ``notifications``: {invalid_items}, "
                "supported options are: \n"
                f"server: {server},\n"
                f"logging: {logging},\n"
                f"project: {project},\n"
                f"nodes: {nodes}"
            )
        args["notifications"] = notifications

    if node_types is not None:
        args["node_types"] = node_types

    return _to_json(
        _gen_server_command(command=_SUBSCRIBE_FOR_PUSH_NOTIFICATIONS, args=args, password=password)
    )


def unlink_registered_file(actor_uid: str, uid: str, password: str = None) -> str:
    """Generate JSON string of ``unlink registered file`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    uid: str
        Uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``unlink registered file`` command.
    """
    args = {}
    args["uid"] = uid

    return _to_json(
        _gen_server_command(
            command=_UNLINK_REGISTERED_FILE, actor_uid=actor_uid, args=args, password=password
        )
    )


def unregister_file(uid: str, password: str = None) -> str:
    """Generate JSON string of ``unregister file`` command.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``unregister file`` command.
    """
    args = {}
    args["uid"] = uid

    return _to_json(_gen_server_command(command=_UNREGISTER_FILE, args=args, password=password))


def unregister_listener(uid: str, password: str = None) -> str:
    """Generate JSON string of ``unregister listener`` command.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``unregister listener`` command.
    """
    args = {}
    args["uid"] = uid

    return _to_json(_gen_server_command(command=_UNREGISTER_LISTENER, args=args, password=password))


def write_monitoring_database(
    actor_uid: str, path: str = None, hid: str = None, password: str = None
) -> str:
    """Generate JSON string of ``write monitoring database`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    path: str, opt
        Path.
    hid: str, opt
        Hid entry.
    password : str, opt
        Password.

    Returns
    -------
    str
        JSON string of ``write monitoring database`` command.
    """
    args = None
    if (path is not None) or (hid is not None):
        args = {}
        if path is not None:
            args["path"] = path
        if hid is not None:
            args["hid"] = hid

    return _to_json(
        _gen_server_command(
            command=_WRITE_MONITORING_DATABASE, actor_uid=actor_uid, args=args, password=password
        )
    )


def _gen_server_command(command, password, args=None, actor_uid=None, hid=None) -> Dict:
    """Generate server command.

    Parameters
    ----------
    command : str
        Command type.
    password : str, opt
        Password.
    args : Dict, opt
        Dictionary with specified arguments.
    actor_uid : str, opt

    Returns
    -------
    scmd : Dict
        Dictionary of server command.

    """
    scmd = {}
    scmd["projects"] = []
    commands = {}
    commands["commands"] = []
    commands["commands"].append(
        _gen_command(command=command, args=args, actor_uid=actor_uid, hid=hid)
    )
    scmd["projects"].append(commands)
    if password:
        scmd["Password"] = password
    return scmd


def _gen_command(command: str, args: Dict = None, actor_uid: str = None, hid: str = None) -> Dict:
    """Generate "commands" for method server command.

    Parameters
    ----------
    type : str
        Command type.
    command : str
        Command name.
    args: dict, optional
        Arguments.
    actor_uid: str, optional
        Actor uid.
    hid: str, optional
        Actor hid.

    Returns
    -------
    dictionary: dict
        Specified dictionary.
    """
    cmd = {}
    cmd["type"] = _builtin
    cmd["command"] = command
    if args:
        cmd["args"] = {}
        for feature, feature_name in args.items():
            cmd["args"][feature] = feature_name
    if actor_uid:
        cmd["actor_uid"] = actor_uid
    if hid:
        cmd["hid"] = hid
    return cmd


def _to_json(dict: Dict) -> str:
    """Convert dictionary to JSON.

    Parameters
    ----------
    dict : Dict
        Dictionary with specified query inputs.

    Returns
    -------
    str
        JSON string.
    """
    return json.dumps(dict, sort_keys=True, ensure_ascii=False)



"""Modify parameters."""
###############################################################################
# (c) 2021 ANSYS, Inc. Unauthorized use, distribution, or duplication is prohibited.
#
# modify flow created by arsm_ten_bar_truss.py
# set lowerbounds of a parameter
# run following command from command line:
# optislang ten_bar_arsm.opf --batch --python ten_bar_modify_parameters.py --no-run

# or choose this file after selecting "Run script file" from Python console's
# context menu
###############################################################################
from py_os_design import *

# import needed modules
from py_os_parameter import *

### ARSM node ###
# get the ARSM node
arsm = find_actor("ARSM")

### Parameter modification ###
# get the parameter manager of ARSM node
params = arsm.parameter_manager
# e.g. SetLowerBound to 5.0 of a parameter
params.set_bounds("area01", PyOSDesignEntry(5.0), PyOSDesignEntry(15.0))
# set parameter manager at ARSM node
arsm.parameter_manager = params



"""Return free ports."""
# (c) 2021 ANSYS, Inc. Unauthorized use, distribution, or duplication is prohibited.
#
# Check for free ports in a specific range (default=49690-49700)

import os
import socket

portrange = range(49690, 49701)

SOCK = None
PORT = 0

for port in portrange:
    try:
        server_address = ("localhost", port)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if not os.name == "posix":
            # set SO_EXCLUSIVEADDRUSE  socket option to ensure that only this socket
            # will listen on the given port
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_EXCLUSIVEADDRUSE, 1)
        sock.bind(server_address)
        sock.listen(5)
        SOCK = sock
        PORT = port
        break
    except Exception as e:
        PORT = 0
        # print('Failed to listen on port {0}: {1}'.format(port, str(e)))

portrange_str = "({0}-{1})".format(portrange[0], portrange[len(portrange) - 1])
if PORT:
    print("Found free port {0} in range {1}".format(PORT, portrange_str))
else:
    print("No free port found in range {0}".format(portrange_str))

if SOCK is not None:
    try:
        SOCK.close()
    except Exception as ex:
        print("Exception closing socket: {0}".format(str(ex)))



"""Contains class ProjectSystem."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Tuple

from ansys.optislang.core.io import RegisteredFile, RegisteredFileUsage
from ansys.optislang.core.nodes import RootSystem

if TYPE_CHECKING:
    from ansys.optislang.core.osl_server import OslServer
    from ansys.optislang.core.project_parametric import (
        CriteriaManager,
        Design,
        ParameterManager,
        ResponseManager,
    )


class Project:
    """Provides the class containing the root system and queries related to the loaded project."""

    def __init__(self, osl_server: OslServer, uid: str) -> None:
        """Initialize an instance of the ``Project`` class.

        Parameters
        ----------
        osl_server: OslServer
            Instance of ``OslServer``.
        uid: str
            Unique ID of the loaded project.
        """
        self.__osl_server = osl_server
        self.__uid = uid
        self.__root_system = RootSystem(
            uid=uid,
            osl_server=self.__osl_server,
        )

    def __str__(self):
        """Return formatted string."""
        return (
            f"Name: {self.get_name()}\n"
            f"Description: {self.get_description()}\n"
            f"Status: {self.get_status()}\n"
            f"Location: {str(self.get_location())}"
        )

    def evaluate_design(self, design: Design, update_design: bool = True) -> Design:
        """Evaluate a design.

        Parameters
        ----------
        design: Design
            Instance of a ``Design`` class with defined parameters.
        update_design: bool, optional
            Determines whether given design should be updated and returned or new instance
            should be created. When ``True`` given design is updated and returned, otherwise
            new ``Design`` is created. Defaults to ``True``.

        Returns
        -------
        Design
            Evaluated design.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.root_system.evaluate_design(design=design, update_design=update_design)

    def get_description(self) -> str:
        """Get the description of the optiSLang project.

        Returns
        -------
        str
            Description of the optiSLang project. If no project is loaded in optiSLang,
            ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.get_project_description()

    def get_location(self) -> Path:
        """Get the path to the optiSLang project file.

        Returns
        -------
        pathlib.Path
            Path to the optiSLang project file. If no project is loaded in the optiSLang,
            ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.get_project_location()

    def get_name(self) -> str:
        """Get the name of the optiSLang project.

        Returns
        -------
        str
            Name of the optiSLang project. If no project is loaded in the optiSLang,
            ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.get_project_name()

    def get_reference_design(self) -> Design:
        """Get a design with reference values of the parameters.

        Returns
        -------
        Design
            Instance of the ``Design`` class with defined parameters and reference values.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.root_system.get_reference_design()

    def get_registered_files(self) -> Tuple[RegisteredFile]:
        """Get all registered files in the current project.

        Returns
        -------
        Tuple[RegisteredFile]
            Tuple with registered files.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        project_registered_files_dicts = self.__osl_server.get_basic_project_info()["projects"][0][
            "registered_files"
        ]
        return tuple(
            [
                RegisteredFile(
                    path=Path(file["local_location"]["split_path"]["head"]),
                    id=file["ident"],
                    comment=file["comment"],
                    tag=file["tag"],
                    usage=file["usage"],
                )
                for file in project_registered_files_dicts
            ]
        )

    def get_result_files(self) -> Tuple[RegisteredFile]:
        """Get result files.

        Returns
        -------
        Tuple[RegisteredFile]
            Tuple with result files

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        registered_files = self.get_registered_files()
        return tuple(
            [file for file in registered_files if file.usage == RegisteredFileUsage.OUTPUT_FILE]
        )

    def get_status(self) -> str:
        """Get the status of the optiSLang project.

        Returns
        -------
        str
            Status of the optiSLang project. If no project is loaded in optiSLang,
            ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__osl_server.get_project_status()

    @property
    def parameter_manager(self) -> ParameterManager:
        """Instance of the ``ParameterManager`` class at the root system.

        Returns
        -------
        ParameterManager
            Parameter manager at the root system.
        """
        return self.__root_system.parameter_manager

    @property
    def response_manager(self) -> ResponseManager:
        """Instance of the ``ResponseManager`` class at the root system.

        Returns
        -------
        ResponseManager
            Response manager at the root system.
        """
        return self.__root_system.response_manager

    @property
    def criteria_manager(self) -> CriteriaManager:
        """Instance of the ``CriteriaManager`` class at the root system.

        Returns
        -------
        CriteriaManager
            Criteria manager at the root system.
        """
        return self.__root_system.criteria_manager

    @property
    def root_system(self) -> RootSystem:
        """Instance of the ``RootSystem`` class.

        Returns
        -------
        RootSystem
            Loaded project's root system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        return self.__root_system

    @property
    def uid(self) -> str:
        """Unique ID of the optiSLang project.

        Returns
        -------
        str
            Unique ID of the loaded project.
        """
        return self.__uid



"""
optiSLang.

core
"""

import os
import sys

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

from ansys.optislang.core.logging import OslLogger

LOG = OslLogger(loglevel="ERROR", log_to_file=False, log_to_stdout=True)
LOG.logger.debug("Loaded logging module as LOG")

__version__ = importlib_metadata.version(__name__.replace(".", "-"))

IRON_PYTHON = sys.platform == "cli"
WINDOWS = os.name == "nt"
LINUX = os.name == "posix"
PY3 = sys.version_info[0] >= 3
# First supported version of optiSLang: 2023R1
FIRST_SUPPORTED_VERSION = 231

from ansys.optislang.core.optislang import Optislang
from ansys.optislang.core.osl_process import OslServerProcess, ServerNotification

# Setup data directory
USER_DATA_PATH = None
LOCAL_DOWNLOADED_EXAMPLES_PATH = None
try:
    import pkgutil

    spec = pkgutil.get_loader("ansys.optislang.core")
    USER_DATA_PATH = os.path.dirname(spec.get_filename())
    if not os.path.exists(USER_DATA_PATH):  # pragma: no cover
        os.makedirs(USER_DATA_PATH)

    LOCAL_DOWNLOADED_EXAMPLES_PATH = os.path.join(USER_DATA_PATH, "examples")
    if not os.path.exists(LOCAL_DOWNLOADED_EXAMPLES_PATH):  # pragma: no cover
        os.makedirs(LOCAL_DOWNLOADED_EXAMPLES_PATH)
    os.environ["OSL_EXAMPLES"] = LOCAL_DOWNLOADED_EXAMPLES_PATH
except:  # pragma: no cover
    pass



"""Contains class which starts and controls local opstiSLang server process."""
from enum import Enum
import logging
import os
from pathlib import Path
import subprocess
import tempfile
from threading import Thread
from typing import Callable, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, Union

import psutil

from ansys.optislang.core import IRON_PYTHON, encoding, utils

if IRON_PYTHON:
    import System


class ServerNotification(Enum):
    """Push notifications available for subscription from the optiSLang server."""

    SERVER_UP = 0
    SERVER_DOWN = 1
    LOG_INFO = 2
    LOG_WARNING = 3
    LOG_ERROR = 4
    LOG_DEBUG = 5
    EXECUTION_STARTED = 6
    PROCESSING_STARTED = 7
    EXECUTION_FINISHED = 8
    NOTHING_PROCESSED = 9
    CHECK_FAILED = 10
    EXEC_FAILED = 11
    ACTOR_STATE_CHANGED = 12
    ACTOR_ACTIVE_CHANGED = 13
    ACTOR_NAME_CHANGED = 14
    ACTOR_CONTENTS_CHANGED = 15
    ACTOR_DATA_CHANGED = 16
    ALL = 17


class OslServerProcess:
    r"""Starts and controls local optiSLang server process.

    Parameters
    ----------
    executable : Union[str, pathlib.Path]
        Path to the optiSLang executable file.
    project_path : Union[str, pathlib.Path], optional
        Path to the optiSLang project file.
        - If the project file exists, it is opened.
        - If the project file does not exist, a new project is created on the specified path.
        - If the path is None, a new project is created in the temporary directory.
        Defaults to ``None``.
    batch : bool, optional
        Determines whether to start optiSLang server in batch mode. Defaults to ``True``.
    port_range : Tuple[int, int], optional
        Defines the port range for optiSLang server. Defaults to ``None``.
    password : str, optional
        The server password. Use when communication with the server requires the request
        to contain a password entry. Defaults to ``None``.
    no_run : bool, optional
        Determines whether not to run the specified project when started in batch mode.
        Defaults to ``None``.

        .. note:: Only supported in batch mode.

    no_save : bool, optional
        Determines whether not to save the specified project after all other actions are completed.
        Defaults to ``False``.

        .. note:: Only supported in batch mode.

    force : bool, optional
        Determines whether to force opening/processing specified project when started in batch mode
        even if issues occur.
        Defaults to ``True``.

        .. note:: Only supported in batch mode.

    reset : bool, optional
        Determines whether to reset specified project after load.
        Defaults to ``False``.

        .. note:: Only supported in batch mode.

    auto_relocate : bool, optional
        Determines whether to automatically relocate missing file paths.
        Defaults to ``False``.

        .. note:: Only supported in batch mode.

    enable_tcp_server : bool, optional
        Determines whether to enable optiSLang TCP server.
        Defaults to ``True``.
    server_info : Union[str, pathlib.Path], optional
        Path to the server information file. If a relative path is provided, it is considered
        to be relative to the project working directory. If ``None``, no server information file
        will be written. Defaults to ``None``.
    log_commands : bool, optional
        Determines whether to display server events in the Message log pane. Defaults to ``False``.
    listener : Tuple[str, int], optional
        Host and port of the remote listener (plain TCP/IP based) to be registered at optiSLang
        server. Defaults to ``None``.
    listener_id : str, optional
        Specific unique ID for the TCP listener. Defaults to ``None``.
    multi_listener : Iterable[Tuple[str, int, Optional[str]]], optional
        Multiple remote listeners (plain TCP/IP based) to be registered at optiSLang server.
        Each listener is a combination of host, port and (optionally) listener ID.
        Defaults to ``None``.
    notifications : Iterable[ServerNotification], optional
        Notifications to be sent to the listener. Defaults to ``None``.
    shutdown_on_finished: bool, optional
        Shut down when execution is finished. Defaults to ``True``.

        .. note:: Only supported in batch mode.

    env_vars : Mapping[str, str], optional
        Additional environmental variables (key and value) for the optiSLang server process.
        Defaults to ``None``.
    logger : Any, optional
        Object for logging. If ``None``, standard logging object is used. Defaults to ``None``.
    log_process_stdout : bool, optional
        Determines whether the process STDOUT is supposed to be logged. Defaults to ``True``.
    log_process_stderr : bool, optional
        Determines whether the process STDERR is supposed to be logged. Defaults to ``True``.
    import_project_properties_file : Union[str, pathlib.Path], optional
        Optional path to a project properties file to import. Defaults to ``None``.
    export_project_properties_file : Union[str, pathlib.Path], optional
        Optional path to a project properties file to export. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    import_placeholders_file : Union[str, pathlib.Path], optional
        Optional path to a placeholders file to import. Defaults to ``None``.
    export_placeholders_file : Union[str, pathlib.Path], optional
        Optional path to a placeholders file to export. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    output_file : Union[str, pathlib.Path], optional
        Optional path to an output file for writing project run results to. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    dump_project_state : Union[str, pathlib.Path], optional
        Optional path to a project state dump file to export. If a relative path is provided,
        it is considered to be relative to the project working directory. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    opx_project_definition_file : Union[str, pathlib.Path], optional
        Optional path to an OPX project definition file. Defaults to ``None``.

        .. note:: Only supported in batch mode.

    additional_args : Iterable[str], optional
        Additional command line arguments used for execution of the optiSLang server process.
        Defaults to ``None``.

    Raises
    ------
    TypeError
        Raised when the optiSLang executable is None.
    FileNotFoundError
        Raised when the optiSLang executable cannot be found.
    ValueError
        Raised when the path to the optiSLang project file is invalid.

    Examples
    --------
    Start optiSLang server process in batch with specified path to the optiSLang project.
    File paths are defined in Windows operating system.

    >>> from ansys.optislang.core import OslServerProcess
    >>> executable = r"C:\Program Files\ANSYS Inc\v221\optiSLang\optislang.com"
    >>> project_path = r"C:\osl_project\project.opf"
    >>> osl_process = OslServerProcess(executable, project_path)
    >>> osl_process.start()
    """

    DEFAULT_PROJECT_FILE = "project.opf"

    def __init__(
        self,
        executable: Union[str, Path] = None,
        project_path: Union[str, Path] = None,
        batch: bool = True,
        port_range: Tuple[int, int] = None,
        password: str = None,
        no_run: bool = None,
        no_save: bool = False,
        force: bool = True,
        reset: bool = False,
        auto_relocate: bool = False,
        enable_tcp_server: bool = True,
        server_info: Union[str, Path] = None,
        log_server_events: bool = False,
        listener: Tuple[str, int] = None,
        listener_id: str = None,
        multi_listener: Iterable[Tuple[str, int, Optional[str]]] = None,
        notifications: Iterable[ServerNotification] = None,
        shutdown_on_finished: bool = True,
        env_vars: Mapping[str, str] = None,
        logger=None,
        log_process_stdout: bool = True,
        log_process_stderr: bool = True,
        import_project_properties_file: Union[str, Path] = None,
        export_project_properties_file: Union[str, Path] = None,
        import_placeholders_file: Union[str, Path] = None,
        export_placeholders_file: Union[str, Path] = None,
        output_file: Union[str, Path] = None,
        dump_project_state: Union[str, Path] = None,
        opx_project_definition_file: Union[str, Path] = None,
        additional_args: Iterable[str] = None,
    ) -> None:
        """Initialize a new instance of the ``OslServerProcess`` class."""
        self._logger = logging.getLogger(__name__) if logger is None else logger
        self.__process = None
        self.__handle_process_output_thread = None

        self.__tempdir = None

        if executable is None:
            osl_exec = utils.get_osl_exec()
            if osl_exec is None:
                raise RuntimeError("No optiSLang executable could be found.")

            osl_version, osl_exec_path = osl_exec
            self.__executable = osl_exec_path
            self._logger.info(
                f"optiSLang executable has been found for version {osl_version} "
                f"on path: {osl_exec_path}"
            )
        elif not os.path.isfile(executable):
            raise FileNotFoundError(f"optiSLang executable cannot be found: {executable}")
        else:
            self.__executable = Path(executable)
        if batch:
            if project_path is None:
                self.__tempdir = tempfile.TemporaryDirectory()
                project_path = Path(self.__tempdir.name) / self.__class__.DEFAULT_PROJECT_FILE

        self.__project_path = self.__class__.__validated_path(project_path, "project_path")

        if self.__project_path.suffix != ".opf":
            raise ValueError("Invalid optiSLang project file.")

        self.__server_info = self.__class__.__validated_path(server_info, "server_info")
        self.__import_project_properties_file = self.__class__.__validated_path(
            import_project_properties_file, "import_project_properties_file"
        )
        self.__export_project_properties_file = self.__class__.__validated_path(
            export_project_properties_file, "export_project_properties_file"
        )
        self.__import_placeholders_file = self.__class__.__validated_path(
            import_placeholders_file, "import_placeholders_file"
        )
        self.__export_placeholders_file = self.__class__.__validated_path(
            export_placeholders_file, "export_placeholders_file"
        )
        self.__output_file = self.__class__.__validated_path(output_file, "output_file")
        self.__dump_project_state = self.__class__.__validated_path(
            dump_project_state, "dump_project_state"
        )
        self.__opx_project_definition_file = self.__class__.__validated_path(
            opx_project_definition_file, "opx_project_definition_file"
        )

        self.__batch = batch
        self.__port_range = port_range
        self.__password = password
        self.__no_run = no_run
        self.__no_save = no_save
        self.__force = force
        self.__reset = reset
        self.__auto_relocate = auto_relocate
        self.__enable_tcp_server = enable_tcp_server
        self.__log_server_events = log_server_events
        self.__listener = listener
        self.__listener_id = listener_id
        self.__multi_listener = multi_listener
        self.__notifications = tuple(notifications) if notifications is not None else None
        self.__shutdown_on_finished = shutdown_on_finished
        self.__env_vars = dict(env_vars) if env_vars is not None else None
        self.__log_process_stdout = log_process_stdout
        self.__log_process_stderr = log_process_stderr
        self.__additional_args = additional_args

    @property
    def executable(self) -> Path:
        """Path to the optiSLang executable file.

        Returns
        -------
        Path
            Path to the optiSLang executable file.
        """
        return self.__executable

    @property
    def project_path(self) -> Path:
        """Path to the optiSLang project file.

        Returns
        -------
        pathlib.Path
            Path to the optiSLang project file.
        """
        return self.__project_path

    @property
    def batch(self) -> bool:
        """Get whether the optiSLang server is supposed to be started in batch mode.

        Returns
        -------
        bool
            ``True`` if the optiSLang server is supposed to be started in batch mode; ``False``
            otherwise.
        """
        return self.__batch

    @property
    def port_range(self) -> Tuple[int, int]:
        """Port range for optiSLang server execution.

        Returns
        -------
        Tuple[int, int]
            Tuple of minimum and maximum port number, if defined; ``None`` otherwise.
        """
        return self.__port_range

    @property
    def password(self) -> str:
        """Server password.

        Returns
        -------
        str
            Server password, if defined; ``None`` otherwise.
        """
        return self.__password

    @property
    def no_run(self) -> bool:
        """Get whether not to run the specified project when started in batch mode .

        Returns
        -------
        bool
            ``True`` if the project is explicitly not supposed to be run;
            ``False`` if the project is explicitly supposed to be run;
            ``None`` otherwise.
        """
        return self.__no_run

    @property
    def no_save(self) -> bool:
        """Get whether not to save the specified project after all other actions are completed.

        Returns
        -------
        bool
            ``True`` if the project is not supposed to be saved after all other actions are
            completed; ``False`` otherwise.
        """
        return self.__no_save

    @property
    def reset(self) -> bool:
        """Get whether to reset specified project after load.

        Returns
        -------
        bool
            ``True`` if project is to be reset after load; ``False`` otherwise.
        """
        return self.__reset

    @property
    def auto_relocate(self) -> bool:
        """Get whether to automatically relocate missing file paths.

        Returns
        -------
        bool
            ``True`` if automatic relocation is enabled; ``False`` otherwise.
        """
        return self.__auto_relocate

    @property
    def enable_tcp_server(self) -> bool:
        """Get whether to enable optiSLang TCP server.

        Returns
        -------
        bool
            ``True`` if optiSLang TCP server is enabled; ``False`` otherwise.
        """
        return self.__enable_tcp_server

    @property
    def server_info(self) -> Union[Path, None]:
        """Path to the server information file.

        Returns
        -------
        Union[pathlib.Path, None]
            Path to the server information file, if defined; ``None`` otherwise.
        """
        return self.__server_info

    @property
    def log_commands(self) -> bool:
        """Get whether to display server events in the Message log pane.

        Returns
        -------
        bool
            ``True`` if server events are supposed to be displayed in the Message log pane;
            ``False`` otherwise.
        """
        return self.__log_server_events

    @property
    def listener(self) -> Tuple[str, int]:
        """Host and port of the remote listener.

        The listener (plain TCP/IP based) is registered at optiSLang server.

        Returns
        -------
        Tuple[str, int]
            Host and port of the listener, if defined; ``None`` otherwise.
        """
        return self.__listener

    @property
    def listener_id(self) -> str:
        """Specific unique ID for the TCP listener.

        Returns
        -------
        str
            Specific unique ID for the TCP listener, if defined; ``None`` otherwise.
        """
        return self.__listener_id

    @property
    def multi_listener(self) -> Iterable[Tuple[str, int, Optional[str]]]:
        """Multi remote listener definitions.

        Aeach listener (plain TCP/IP based) is registered at optiSLang server.

        Returns
        -------
        Iterable[Tuple[str, int, Optional[str]]]
            Multi remote listener combinations, if defined; ``None`` otherwise.
        """
        return self.__multi_listener

    @property
    def notifications(self) -> Tuple[ServerNotification, ...]:
        """Notifications which are sent to the listener.

        Returns
        -------
        Tuple[ServerNotification,...]
            Notifications which are sent to the listener, if defined; ``None`` otherwise.
        """
        return self.__notifications

    @property
    def env_vars(self) -> Dict[str, str]:
        """Additional environmental variables for the optiSLang server process.

        Returns
        -------
        Dict[str, str]
            Dictionary of additional environmental variables, if defined; ``None`` otherwise.
        """
        return self.__env_vars

    @property
    def logger(self):
        """Object used for logging.

        Returns
        -------
        :class:
            Logger object.
        """
        return self._logger

    @property
    def log_process_stdout(self) -> bool:
        """Get whether the STDOUT of the optiSLang server process is to be logged.

        Returns
        -------
        bool
            ``True`` if the STDOUT of the optiSLang server process is to be logged; ``False``
            otherwise.
        """
        return self.__log_process_stdout

    @property
    def log_process_stderr(self) -> bool:
        """Get whether the STDERR of the optiSLang server process is to be logged.

        Returns
        -------
        bool
            ``True`` if the STDERR of the optiSLang server process is to be logged; ``False``
            otherwise.
        """
        return self.__log_process_stderr

    @property
    def additional_args(self) -> Tuple[str, ...]:
        """Additional command line arguments used for optiSLang server process execution.

        Returns
        -------
        Tuple[str, ...]
            Additional command line arguments, if defined; ``None`` otherwise.
        """
        return self.__additional_args

    @property
    def pid(self) -> Union[int, None]:
        """Process ID.

        Returns
        -------
        Union[int, None]
            Process ID, if exists; ``None`` otherwise.
        """
        return None if self.__process is None else self.__process.pid

    @property
    def shutdown_on_finished(self) -> str:
        """Whether to shut down when execution is finished.

        Returns
        -------
        str
            Whether to shut down when execution is finished.
        """
        return self.__shutdown_on_finished

    @property
    def import_project_properties_file(self) -> Union[Path, None]:
        """Path to the project properties import file.

        Returns
        -------
        Union[pathlib.Path, None]
            Path to the project properties import file, if defined; ``None`` otherwise.
        """
        return self.__import_project_properties_file

    @property
    def export_project_properties_file(self) -> Union[Path, None]:
        """Path to the project properties export file.

        Returns
        -------
        Union[pathlib.Path, None]
            Path to the project properties export file, if defined; ``None`` otherwise.
        """
        return self.__export_project_properties_file

    @property
    def import_placeholders_file(self) -> Union[Path, None]:
        """Path to the placeholders import file.

        Returns
        -------
        Union[pathlib.Path, None]
            Path to the placeholders import file, if defined; ``None`` otherwise.
        """
        return self.__import_placeholders_file

    @property
    def export_placeholders_file(self) -> Union[Path, None]:
        """Path to the placeholders export file.

        Returns
        -------
        Union[pathlib.Path, None]
            Path to the placeholders export file, if defined; ``None`` otherwise.
        """
        return self.__export_placeholders_file

    @property
    def output_file(self) -> Union[Path, None]:
        """Path to the output file for writing project run results to.

        Returns
        -------
        Union[pathlib.Path, None]
            Path to the output file for writing project run results to, if defined;
            ``None`` otherwise.
        """
        return self.__output_file

    @property
    def dump_project_state(self) -> Union[Path, None]:
        """Path to a project state dump file to export.

        Returns
        -------
        Union[pathlib.Path, None]
            Path to a project state dump file to export, if defined; ``None`` otherwise.
        """
        return self.__dump_project_state

    @property
    def opx_project_definition_file(self) -> Union[Path, None]:
        """Path to the OPX project definition file.

        Returns
        -------
        Union[pathlib.Path, None]
            Path to the OPX project definition file, if defined;
            ``None`` otherwise.
        """
        return self.__opx_project_definition_file

    def __enter__(self):
        """Enter the context."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Exit the context."""
        self.terminate()

    def _get_process_args(self) -> List[str]:
        """Get the command line arguments for the optiSLang server process execution.

        Returns
        -------
        List[str]
            List of command line arguments.
        """
        args = [str(self.__executable)]

        if self.__batch:
            args.append("-b")  # Start batch mode

        if self.__project_path:
            if not self.__project_path.is_file():
                # Creates a new project in the provided path.
                if IRON_PYTHON:
                    args.append(f'--new="{str(self.__project_path)}"')
                else:
                    args.append(f"--new={str(self.__project_path)}")
            else:
                # Opens existing project
                if IRON_PYTHON:
                    args.append(f'"{str(self.__project_path)}"')
                else:
                    args.append(str(self.__project_path))

        if self.__batch:
            if self.__opx_project_definition_file:
                if IRON_PYTHON:
                    args.append("--python")
                    args.append(f'"{str(utils.get_osl_opx_import_script(self.__executable))}"')
                    args.append("--script-arg")
                    args.append(f'"{str(self.__opx_project_definition_file)}"')
                else:
                    args.append("--python")
                    args.append(str(utils.get_osl_opx_import_script(self.__executable)))
                    args.append("--script-arg")
                    args.append(str(self.__opx_project_definition_file))
            if self.__no_run is None or self.__no_run:
                # Do not run the specified projects.
                args.append("--no-run")
            if self.__no_save:
                # Do not save the specified projects after all other actions have been completed.
                args.append("--no-save")
            if self.__force:
                # Forces projects to be processed, even if they are incomplete
                # (can be used for damaged projects).
                args.append("--force")
            if self.__shutdown_on_finished:
                args.append("--shutdown-on-finished")
            if self.__reset:
                args.append("--reset")
            if self.__auto_relocate:
                args.append("--autorelocate")
            if self.__export_project_properties_file is not None:
                args.append("--export-project-properties")
                if IRON_PYTHON:
                    args.append(f'"{str(self.__export_project_properties_file)}"')
                else:
                    args.append(str(self.__export_project_properties_file))
            if self.__export_placeholders_file is not None:
                args.append("--export-values")
                if IRON_PYTHON:
                    args.append(f'"{str(self.__export_placeholders_file)}"')
                else:
                    args.append(str(self.__export_placeholders_file))
            if self.__output_file is not None:
                args.append("--output-file")
                if IRON_PYTHON:
                    args.append(f'"{str(self.__output_file)}"')
                else:
                    args.append(str(self.__output_file))
            if self.__dump_project_state is not None:
                if IRON_PYTHON:
                    args.append(f'--dump-project-state="{str(self.__dump_project_state)}"')
                else:
                    args.append(f"--dump-project-state={str(self.__dump_project_state)}")

        # Enables remote surveillance (plain TCP/IP based), the port indication is optional.
        if self.__enable_tcp_server:
            if self.__port_range is not None:
                args.append(f"--enable-tcp-server={self.__port_range[0]}-{self.__port_range[1]}")
            else:
                args.append("--enable-tcp-server")

        if self.__password is not None:
            # Submits the server password. Use when communication with the server requires
            # the request to contain a password entry.
            args.append(f"--server-password={self.__password}")

        if self.__server_info is not None:
            # Writes the server information file using the file path specified. If an absolute path
            # is not supplied, it is considered to be relative to the project working directory.
            if IRON_PYTHON:
                args.append(f'--write-server-info="{str(self.__server_info)}"')
            else:
                args.append(f"--write-server-info={str(self.__server_info)}")

        if self.__log_server_events:
            # Displays server events in the Message log pane.
            args.append("--log-server-events")

        if self.__listener is not None:
            # Registers the remote listener (plain TCP/IP based) for specified host and port.
            args.append(f"--register-tcp-listener={self.__listener[0]}:{self.__listener[1]}")

        if self.__listener_id is not None:
            # Sets a specific unique ID for the TCP listener.
            args.append(f"--tcp-listener-id={self.__listener_id}")

        if self.__multi_listener is not None:
            if len(self.__multi_listener) >= 2:
                args.append("--register-multi-tcp-listeners")
            for listener in self.__multi_listener:
                if len(listener) >= 3 and listener[2] is not None:
                    args.append(f"{listener[0]}:{listener[1]}+{listener[2]}")
                else:
                    args.append(f"{listener[0]}:{listener[1]}")

        if self.__notifications is not None:
            # Subscribe to push notifications sent to the listener.
            args.append("--enable-notifications")
            for notification in self.__notifications:
                args.append(notification.name)

        if self.__import_project_properties_file is not None:
            args.append("--import-project-properties")
            if IRON_PYTHON:
                args.append(f'"{str(self.__import_project_properties_file)}"')
            else:
                args.append(str(self.__import_project_properties_file))

        if self.__import_placeholders_file is not None:
            args.append("--import-values")
            if IRON_PYTHON:
                args.append(f'"{str(self.__import_placeholders_file)}"')
            else:
                args.append(str(self.__import_placeholders_file))

        if self.__additional_args is not None:
            for arg in self.__additional_args:
                args.append(arg)

        return args

    def __remove_server_info_files(self):
        """Remove server information files.

        Remove all .ini files located in the same directory as optiSLang project file.
        """
        if self.__project_path is not None:
            project_dir = os.path.dirname(self.__project_path)
            for file in os.listdir(project_dir):
                if file.endswith(".ini"):
                    info_file = os.path.join(project_dir, file)
                    os.remove(info_file)
                    self._logger.info("The server information file %s has been removed", info_file)

    def start(self, remove_ini_files: bool = True):
        """Start new optiSLang server process.

        Parameters
        ----------
        remove_ini_files : bool, optional
            Determines whether to remove all server information files (.ini) located in the same
            directory as the optiSLang project file. Defaults to ``True``.

        Raises
        ------
        RuntimeError
            Raised when the process is already running.
        """
        if self.is_running():
            raise RuntimeError("Process is running.")

        if remove_ini_files:
            self.__remove_server_info_files()

        args = self._get_process_args()
        env_vars = os.environ.copy()
        if self.__env_vars is not None:
            env_vars.update(self.__env_vars)

        if IRON_PYTHON:
            self.__start_in_iron_python(args, env_vars)
        else:
            self.__start_in_python(args, env_vars)

        if self.__log_process_stdout or self.__log_process_stderr:
            self.__start_process_output_thread()

    def __start_in_iron_python(self, args: Sequence[str], env_vars: Mapping[str, str]):
        """Start new optiSLang server process with IronPython interpreter.

        Parameters
        ----------
        args : Sequence[str]
            Sequence of command line arguments.

        env_vars : Mapping[str, str]
            Environment variables.
        """
        start_info = System.Diagnostics.ProcessStartInfo()

        if env_vars is not None:
            for var_name, var_value in env_vars.items():
                var_value = var_value if var_value is not None else ""  # FIX: This is not necessary
                if not start_info.EnvironmentVariables.ContainsKey(var_name):
                    start_info.EnvironmentVariables.Add(var_name, var_value)

        # TODO: Add comment why this is done
        if start_info.EnvironmentVariables.ContainsKey("LD_LIBRARY_PATH"):
            start_info.EnvironmentVariables.Remove("LD_LIBRARY_PATH")
            self._logger.debug(
                "LD_LIBRARY_PATH environment variable has been removed before "
                "start of the optiSLang process."
            )

        start_info.UseShellExecute = False
        start_info.RedirectStandardError = True
        start_info.RedirectStandardOutput = True
        start_info.CreateNoWindow = True
        start_info.FileName = self.executable
        start_info.Arguments = " ".join(args)

        self.__process = System.Diagnostics.Process()
        self.__process.StartInfo = start_info

        self._logger.debug("Executing process %s", args)
        self.__process.Start()

    def __start_in_python(self, args: Sequence[str], env_vars: Mapping[str, str]):
        """Start new optiSLang server process with Python interpreter.

        Parameters
        ----------
        args : Sequence[str]
            Sequence of command line arguments.

        env_vars : Mapping[str, str]
            Environment variables.
        """
        for i in range(len(args)):
            args[i] = encoding.to_ascii_safe(args[i])

        for var_name in env_vars:
            env_vars[var_name] = encoding.to_ascii_safe(env_vars[var_name])

        # TODO: Add comment why this is done
        if "LD_LIBRARY_PATH" in env_vars:
            env_vars.pop("LD_LIBRARY_PATH")
            self._logger.debug(
                "LD_LIBRARY_PATH environment variable has been removed before "
                "start of the optiSLang process."
            )

        self._logger.debug("Executing process %s", args)
        self.__process = subprocess.Popen(
            args,
            env=env_vars,
            cwd=os.getcwd(),
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            shell=False,
        )
        self._logger.debug("optiSLang server process has started with PID: %d", self.__process.pid)

    def __terminate_osl_child_processes(self):
        """Terminate all child processes of the optiSLang server process."""
        try:
            parent = psutil.Process(self.__process.pid)
        except psutil.NoSuchProcess:
            return

        children = parent.children(recursive=True)
        for process in children:
            try:
                process.terminate()
            except psutil.NoSuchProcess:
                self._logger.debug(
                    f"Cannot terminate child process PID: {process.pid}. "
                    "The process does not exist."
                )

        gone, alive = psutil.wait_procs(children, timeout=3)

        for process in alive:
            self._logger.debug(
                f"optiSLang server child process {process} could not be terminated "
                "and will be killed.",
            )
            process.kill()

    def terminate(self):
        """Terminate optiSLang server process."""
        if self.__process is not None:
            self.__terminate_osl_child_processes()
            self.__process.terminate()

        if (
            self.__handle_process_output_thread is not None
            and self.__handle_process_output_thread.is_alive()
        ):
            self.__handle_process_output_thread.join()
        self.__handle_process_output_thread = None

        if self.__tempdir is not None:
            self.__tempdir.cleanup()
            self.__tempdir = None

    def is_running(self) -> bool:
        """Determine whether the optiSLang server process is running.

        Returns
        -------
        bool
            ``True`` if the optiSLang server process is running; ``False`` otherwise.
        """
        if self.__process is None:
            return False

        return self.__process.poll() is None

    def wait_for_finished(self):
        """Wait for the process to finish."""
        if self.__process is not None and self.is_running():
            self.__process.wait()

    def __start_process_output_thread(self):
        """Start new thread responsible for logging of STDOUT/STDERR of the optiSLang process."""

        def finalize_process(process, **kwargs):
            process.wait(**kwargs)

        self.__handle_process_output_thread = Thread(
            target=self.__class__.__handle_process_output,
            name="PyOptiSLang.ProcessOutputHandlerThread",
            args=(
                self.__process,
                self._logger.debug if self.__log_process_stdout else None,
                self._logger.warning if self.__log_process_stderr else None,
                finalize_process,
                True,
                self._logger,
            ),
            daemon=True,
        )
        self.__handle_process_output_thread.start()

    @staticmethod
    def __handle_process_output(
        process: subprocess.Popen,
        stdout_handler: Callable[[str], None],
        stderr_handler: Callable[[str], None],
        finalizer=None,
        decode_streams: bool = True,
        logger=None,
    ):
        """Handle STDOUT/STDERR of the specified process.

        Registers for notifications to lean that process output is ready to read, and dispatches
        lines to the respective line handlers. This function returns once the finalizer returns.

        [the code was taken from gitpython]

        Parameters
        ----------
        process : subprocess.Popen
            Process which STDOUT or STDERR is supposed to be handled.

        stdout_handler : Callable[[str], None], None
            Handler for STDOUT.

        stderr_handler : Callable[[str], None], None
            Handler for STDERR. It is supposed to be a function with one argument of str.

        finalizer : Callable[[subprocess.Popen,...], Any], optional
            Function which finalizes output process handling. Defaults to ``None``.

        decode_streams : bool, optional
            Determines whether to safely decode STDOUT/STDERR streams before pushing their
            contents to handlers. Should be set to ``False`` if 'universal_newline == True'
            (then streams are in text-mode) or if decoding must happen later. Defaults to ``True``.

        logger : Any, optional
            Object for logging. Defaults to ``None``.

        Returns
        -------
        Any
            Result of finalizer.
        """
        logger.debug("Start to handling optiSLang server process output.")

        # Use 2 "pupm" threads and wait for both to finish.
        if IRON_PYTHON:

            def stream_reader(cmdline, name, stream, is_decode, handler):
                try:
                    while True:
                        if stream.EndOfStream:
                            break
                        line = stream.ReadLine()
                        if handler:
                            try:
                                if is_decode:
                                    line = encoding.force_text(line)
                                handler("optiSLang " + name + ": " + line)
                            except:
                                handler("optiSLang " + name + ": " + line)
                except Exception as ex:
                    if logger is not None:
                        logger.debug(
                            "Pumping {0} of cmd({1}) failed due to: {2}".format(name, cmdline, ex)
                        )
                finally:
                    stream.Close()

        else:

            def stream_reader(cmdline, name, stream, is_decode, handler):
                try:
                    for line in stream:
                        if handler:
                            try:
                                if is_decode:
                                    line = encoding.force_text(line)
                                handler("optiSLang " + name + ": " + line)
                            except:
                                handler("optiSLang " + name + ": " + line)
                except Exception as ex:
                    if logger is not None:
                        logger.debug(
                            "Pumping {0} of cmd({1}) failed due to: {2}".format(name, cmdline, ex)
                        )
                finally:
                    stream.close()

        cmdline = getattr(process, "args", "")  # PY3+ only
        if not isinstance(cmdline, (tuple, list)):
            cmdline = cmdline.split()

        pumps = []
        if process.stdout and stdout_handler is not None:
            pumps.append(("Stdout", process.stdout, stdout_handler))
        if process.stderr and stderr_handler is not None:
            pumps.append(("Stderr", process.stderr, stderr_handler))

        threads = []

        for name, stream, handler in pumps:
            thread = Thread(
                target=stream_reader,
                name="PyOptiSLang." + name + "HandlerThread",
                args=(cmdline, name, stream, decode_streams, handler),
            )
            thread.daemon = True
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()

        if finalizer:
            return finalizer(process)

    @staticmethod
    def __validated_path(
        path: Union[str, Path],
        path_arg_name: str,
    ):
        """Validate a path.

        Parameters
        ----------
        path : Union[str, pathlib.Path]
            The path to validate.
        path_arg_name : str
            Path argument name/description.

        Returns
        -------
        pathlib.Path
            Validated path.
        """
        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path) and path is not None:
            raise TypeError(
                f"Invalid type of {path_arg_name}: {type(path)},"
                "Union[str, pathlib.Path] is supported."
            )

        return path
